"""Asana source helpers"""

from asana import Client as AsanaClient


def get_client(
    access_token: str,
) -> AsanaClient:
    """
    Returns an Asana API client.
    Args:
        access_token (str): The access token to authenticate the Asana API client.
    Returns:
        AsanaClient: The Asana API client.
    """
    return AsanaClient.access_token(access_token)



from typing import Iterable

import dlt
import smartsheet  # type: ignore
from dlt.extract import DltResource


@dlt.source
def smartsheet_source(
    access_token: str,
    sheet_id: str,
) -> Iterable[DltResource]:
    """
    A DLT source for Smartsheet.

    Args:
        access_token: The Smartsheet API access token.
        sheet_id: The ID of the sheet to load.

    Returns:
        An iterable of DLT resources.
    """

    # Initialize Smartsheet client
    smartsheet_client = smartsheet.Smartsheet(access_token)
    smartsheet_client.errors_as_exceptions(True)

    # The SDK expects sheet_id to be an int
    sheet_id_int = int(sheet_id)
    # Sanitize the sheet name to be a valid resource name
    # We get objectValue to ensure `name` attribute is populated for the sheet
    sheet_details = smartsheet_client.Sheets.get_sheet(
        sheet_id_int, include=["objectValue"]
    )
    sheet_name = sheet_details.name
    resource_name = f"sheet_{sheet_name.replace(' ', '_').lower()}"

    yield dlt.resource(
        _get_sheet_data(smartsheet_client, sheet_id_int),
        name=resource_name,
        write_disposition="replace",
    )


def _get_sheet_data(smartsheet_client: smartsheet.Smartsheet, sheet_id: int):
    """Helper function to get all rows from a sheet."""
    sheet = smartsheet_client.Sheets.get_sheet(sheet_id)
    # Transform rows to a list of dictionaries
    column_titles = [col.title for col in sheet.columns]
    for row in sheet.rows:
        row_data = {}
        for i, cell in enumerate(row.cells):
            row_data[column_titles[i]] = cell.value
        yield row_data



from typing import Iterable

import dlt
import pendulum
from dlt.common.typing import TDataItem
from dlt.sources import DltResource

from .client import InfluxClient

@dlt.source(max_table_nesting=0)
def influxdb_source(
    measurement: str,
    host: str,
    org: str,
    bucket: str,
    token: str = dlt.secrets.value,
    secure: bool = True,
    start_date: pendulum.DateTime = pendulum.datetime(2024, 1, 1),
    end_date: pendulum.DateTime | None = None,
) -> Iterable[DltResource]:
    client = InfluxClient(
        url=host, token=token, org=org, bucket=bucket, verify_ssl=secure
    )

    @dlt.resource(name=measurement)
    def fetch_table(
        timestamp=dlt.sources.incremental(
            "time",
            initial_value=start_date,
            end_value=end_date,
            range_start="closed",
            range_end="closed",
        ),
    ) -> Iterable[TDataItem]:
        if timestamp.last_value is None:
            start = start_date.isoformat()
        else:
            start = timestamp.last_value.isoformat()
        if timestamp.end_value is None:
            end = pendulum.now().isoformat()
        else:
            end = timestamp.end_value.isoformat()
        yield from client.fetch_measurement(measurement, start, end)

    return fetch_table



from typing import Iterable

import dlt
import pendulum
from dlt.common.typing import TDataItem
from dlt.sources import DltResource
from pendulum import Date

from .dimension_time_enum import Dimension, TimeGranularity
from .helpers import LinkedInAdsAPI, find_intervals


@dlt.source(max_table_nesting=0)
def linked_in_ads_source(
    start_date: Date,
    end_date: Date | None,
    access_token: str,
    account_ids: list[str],
    dimension: Dimension,
    metrics: list[str],
    time_granularity: TimeGranularity,
) -> DltResource:
    linkedin_api = LinkedInAdsAPI(
        access_token=access_token,
        account_ids=account_ids,
        dimension=dimension,
        metrics=metrics,
        time_granularity=time_granularity,
    )

    if time_granularity == TimeGranularity.daily:
        primary_key = [dimension.value, "date"]
        incremental_loading_param = "date"
    else:
        primary_key = [dimension.value, "start_date", "end_date"]
        incremental_loading_param = "start_date"

    @dlt.resource(write_disposition="merge", primary_key=primary_key)
    def custom_reports(
        dateTime=(
            dlt.sources.incremental(
                incremental_loading_param,
                initial_value=start_date,
                end_value=end_date,
                range_start="closed",
                range_end="closed",
            )
        ),
    ) -> Iterable[TDataItem]:
        if dateTime.end_value is None:
            end_date = pendulum.now().date()
        else:
            end_date = dateTime.end_value

        list_of_interval = find_intervals(
            start_date=dateTime.last_value,
            end_date=end_date,
            time_granularity=time_granularity,
        )
        for start, end in list_of_interval:
            yield linkedin_api.fetch_pages(start, end)

    return custom_reports



from typing import Any, Dict, Iterable, Sequence

import dlt
import pendulum
from dlt.common.typing import TAnyDateTime, TDataItem
from dlt.sources import DltResource

from .helpers import ZoomClient


@dlt.source(name="zoom", max_table_nesting=0)
def zoom_source(
    client_id: str,
    client_secret: str,
    account_id: str,
    start_date: pendulum.DateTime,
    end_date: pendulum.DateTime | None = None,
) -> Sequence[DltResource]:
    """Create a Zoom source with meetings resource for all users in the account."""
    client = ZoomClient(
        client_id=client_id,
        client_secret=client_secret,
        account_id=account_id,
    )

    @dlt.resource(write_disposition="merge", primary_key="id")
    def meetings(
        datetime: dlt.sources.incremental[TAnyDateTime] = dlt.sources.incremental(
            "start_time",
            initial_value=start_date.isoformat(),
            end_value=end_date.isoformat() if end_date is not None else None,
            range_start="closed",
            range_end="closed",
        ),
    ) -> Iterable[TDataItem]:
        if datetime.last_value:
            start_dt = pendulum.parse(datetime.last_value)
        else:
            start_dt = pendulum.parse(start_date)

        if end_date is None:
            end_dt = pendulum.now("UTC")
        else:
            end_dt = pendulum.parse(datetime.end_value)

        base_params: Dict[str, Any] = {
            "type": "scheduled",
            "page_size": 300,
            "from": start_dt.to_date_string(),
            "to": end_dt.to_date_string(),
        }

        for user in client.get_users():
            user_id = user["id"]
            yield from client.get_meetings(user_id, base_params)

    @dlt.resource(write_disposition="merge", primary_key="id")
    def users() -> Iterable[TDataItem]:
        yield from client.get_users()

    @dlt.resource(write_disposition="merge", primary_key="id")
    def participants(
        datetime: dlt.sources.incremental[TAnyDateTime] = dlt.sources.incremental(
            "join_time",
            initial_value=start_date.isoformat(),
            end_value=end_date.isoformat() if end_date is not None else None,
            range_start="closed",
            range_end="closed",
        ),
    ) -> Iterable[TDataItem]:
        if datetime.last_value:
            start_dt = pendulum.parse(datetime.last_value)
        else:
            start_dt = pendulum.parse(start_date)

        if end_date is None:
            end_dt = pendulum.now("UTC")
        else:
            end_dt = pendulum.parse(datetime.end_value)

        participant_params: Dict[str, Any] = {
            "page_size": 300,
        }
        meeting_params = {
            "type": "previous_meetings",
            "page_size": 300,
        }
        for user in client.get_users():
            user_id = user["id"]
            for meeting in client.get_meetings(user_id=user_id, params=meeting_params):
                meeting_id = meeting["id"]
                yield from client.get_participants(
                    meeting_id=meeting_id,
                    params=participant_params,
                    start_date=start_dt,
                    end_date=end_dt,
                )

    return meetings, users, participants



"""Zendesk source settings and constants"""

from dlt.common import pendulum

DEFAULT_START_DATE = pendulum.datetime(year=2024, month=10, day=3)

INCREMENTAL_PAGE_SIZE = 1000
PAGE_SIZE = 100


CUSTOM_FIELDS_STATE_KEY = "ticket_custom_fields_v2"

# Tuples of (Resource name, endpoint URL, data_key, supports pagination)
# data_key is the key which data list is nested under in responses
# if the data key is None it is assumed to be the same as the resource name
# The last element of the tuple says if endpoint supports cursor pagination
SUPPORT_ENDPOINTS = [
    ("users", "/api/v2/users.json", "users", True),
    ("sla_policies", "/api/v2/slas/policies.json", None, False),
    ("groups", "/api/v2/groups.json", None, True),
    ("organizations", "/api/v2/organizations.json", None, True),
    ("brands", "/api/v2/brands.json", None, True),
]

SUPPORT_EXTRA_ENDPOINTS = [
    ("activities", "/api/v2/activities.json", None, True),
    ("automations", "/api/v2/automations.json", None, True),
    ("macros", "/api/v2/macros.json", None, True),
    ("recipient_addresses", "/api/v2/recipient_addresses.json", None, True),
    ("requests", "/api/v2/requests.json", None, True),
    ("targets", "/api/v2/targets.json", None, False),
    ("ticket_forms", "/api/v2/ticket_forms.json", None, False),
    ("ticket_metrics", "/api/v2/ticket_metrics.json", None, True),
    ("triggers", "/api/v2/triggers.json", None, True),
    ("user_fields", "/api/v2/user_fields.json", None, True),
]

TALK_ENDPOINTS = [
    ("calls", "/api/v2/channels/voice/calls", None, False),
    ("addresses", "/api/v2/channels/voice/addresses", None, False),
    ("greetings", "/api/v2/channels/voice/greetings", None, False),
    ("phone_numbers", "/api/v2/channels/voice/phone_numbers", None, False),
    ("settings", "/api/v2/channels/voice/settings", None, False),
    ("lines", "/api/v2/channels/voice/lines", None, False),
    ("agents_activity", "/api/v2/channels/voice/stats/agents_activity", None, False),
    (
        "current_queue_activity",
        "/api/v2/channels/voice/stats/current_queue_activity",
        None,
        False,
    ),
]

INCREMENTAL_TALK_ENDPOINTS = {
    "calls": "/api/v2/channels/voice/stats/incremental/calls.json",
    "legs": "/api/v2/channels/voice/stats/incremental/legs.json",
}



"""Stripe analytics source settings and constants"""

# the most popular endpoints
# Full list of the Stripe API endpoints you can find here: https://stripe.com/docs/api.
ENDPOINTS = {
    "account": "Account",
    "applepaydomain": "ApplePayDomain",
    "apple_pay_domain": "ApplePayDomain",
    "applicationfee": "ApplicationFee",
    "application_fee": "ApplicationFee",
    "checkoutsession": "CheckoutSession",
    "checkout_session": "CheckoutSession",
    "coupon": "Coupon",
    "charge": "Charge",
    "customer": "Customer",
    "dispute": "Dispute",
    "paymentintent": "PaymentIntent",
    "payment_intent": "PaymentIntent",
    "paymentlink": "PaymentLink",
    "payment_link": "PaymentLink",
    "paymentmethod": "PaymentMethod",
    "payment_method": "PaymentMethod",
    "paymentmethoddomain": "PaymentMethodDomain",
    "payment_method_domain": "PaymentMethodDomain",
    "payout": "Payout",
    "plan": "Plan",
    "price": "Price",
    "product": "Product",
    "promotioncode": "PromotionCode",
    "promotion_code": "PromotionCode",
    "quote": "Quote",
    "refund": "Refund",
    "review": "Review",
    "setupattempt": "SetupAttempt",
    "setup_attempt": "SetupAttempt",
    "setupintent": "SetupIntent",
    "setup_intent": "SetupIntent",
    "shippingrate": "ShippingRate",
    "shipping_rate": "ShippingRate",
    "subscription": "Subscription",
    "subscriptionitem": "SubscriptionItem",
    "subscription_item": "SubscriptionItem",
    "subscriptionschedule": "SubscriptionSchedule",
    "subscription_schedule": "SubscriptionSchedule",
    "transfer": "Transfer",
    "taxcode": "TaxCode",
    "tax_code": "TaxCode",
    "taxid": "TaxId",
    "tax_id": "TaxId",
    "taxrate": "TaxRate",
    "tax_rate": "TaxRate",
    "topup": "Topup",
    "top_up": "Topup",
    "webhookendpoint": "WebhookEndpoint",
    "webhook_endpoint": "WebhookEndpoint",
    "invoice": "Invoice",
    "invoiceitem": "InvoiceItem",
    "invoice_item": "InvoiceItem",
    "invoicelineitem": "InvoiceLineItem",
    "invoice_line_item": "InvoiceLineItem",
    "balancetransaction": "BalanceTransaction",
    "balance_transaction": "BalanceTransaction",
    "creditnote": "CreditNote",
    "credit_note": "CreditNote",
    "event": "Event",
}



from typing import Optional

import requests
from dlt.sources.helpers.requests import Client
from requests.exceptions import HTTPError


class AppsflyerClient:
    def __init__(self, api_key: str):
        self.api_key = api_key
        self.uri = "https://hq1.appsflyer.com/api/master-agg-data/v4/app/all"

    def __get_headers(self):
        return {
            "Authorization": f"{self.api_key}",
            "accept": "text/json",
        }

    def _fetch_data(
        self,
        from_date: str,
        to_date: str,
        dimensions: list[str],
        metrics: list[str],
        maximum_rows=1000000,
    ):
        excluded_metrics = exclude_metrics_for_date_range(metrics, from_date, to_date)
        included_metrics = [
            metric for metric in metrics if metric not in excluded_metrics
        ]

        params = {
            "from": from_date,
            "to": to_date,
            "groupings": ",".join(dimensions),
            "kpis": ",".join(included_metrics),
            "format": "json",
            "maximum_rows": maximum_rows,
        }

        def retry_on_limit(
            response: Optional[requests.Response], exception: Optional[BaseException]
        ) -> bool:
            return (
                isinstance(response, requests.Response) and response.status_code == 429
            )

        request_client = Client(
            raise_for_status=False,
            retry_condition=retry_on_limit,
            request_max_attempts=12,
            request_backoff_factor=2,
        ).session

        try:
            response = request_client.get(
                url=self.uri, headers=self.__get_headers(), params=params
            )

            if response.status_code == 200:
                result = response.json()
                yield standardize_keys(result, excluded_metrics)
            else:
                raise HTTPError(
                    f"Request failed with status code: {response.status_code}: {response.text}"
                )

        except requests.RequestException as e:
            raise HTTPError(f"Request failed: {e}")


def standardize_keys(data: list[dict], excluded_metrics: list[str]) -> list[dict]:
    def fix_key(key: str) -> str:
        return key.lower().replace("-", "").replace("  ", "_").replace(" ", "_")

    standardized = []
    for item in data:
        standardized_item = {}
        for key, value in item.items():
            standardized_item[fix_key(key)] = value

        for metric in excluded_metrics:
            if metric not in standardized_item:
                standardized_item[fix_key(metric)] = None

        standardized.append(standardized_item)

    return standardized


def exclude_metrics_for_date_range(
    metrics: list[str], from_date: str, to_date: str
) -> list[str]:
    """
    Some of the cohort metrics are not available if there hasn't been enough time to have data for that cohort.
    This means if you request data for yesterday with cohort day 7 metrics, you will get an error because 7 days hasn't passed yet.
    One would expect the API to handle this gracefully, but it doesn't.

    This function will exclude the metrics that are not available for the given date range.
    """
    import pendulum

    excluded_metrics = []
    days_between_today_and_end = (pendulum.now() - pendulum.parse(to_date)).days  # type: ignore
    for metric in metrics:
        if "cohort_day_" in metric:
            day_count = int(metric.split("_")[2])
            if days_between_today_and_end <= day_count:
                excluded_metrics.append(metric)
    return excluded_metrics



from typing import Any, Dict, Iterator, Optional

import requests

LINEAR_GRAPHQL_ENDPOINT = "https://api.linear.app/graphql"


def _graphql(
    api_key: str, query: str, variables: Optional[Dict[str, Any]] = None
) -> Dict[str, Any]:
    headers = {"Authorization": api_key, "Content-Type": "application/json"}
    response = requests.post(
        LINEAR_GRAPHQL_ENDPOINT,
        json={"query": query, "variables": variables or {}},
        headers=headers,
    )
    response.raise_for_status()
    payload = response.json()
    if "errors" in payload:
        raise ValueError(str(payload["errors"]))
    return payload["data"]


def _paginate(api_key: str, query: str, root: str) -> Iterator[Dict[str, Any]]:
    cursor: Optional[str] = None
    while True:
        data = _graphql(api_key, query, {"cursor": cursor})[root]
        for item in data["nodes"]:
            yield item
        if not data["pageInfo"]["hasNextPage"]:
            break
        cursor = data["pageInfo"]["endCursor"]


def _normalize_issue(item: Dict[str, Any]) -> Dict[str, Any]:
    field_mapping = {
        "assignee": "assignee_id",
        "creator": "creator_id",
        "state": "state_id",
        "cycle": "cycle_id",
        "project": "project_id",
    }
    for key, value in field_mapping.items():
        if item.get(key):
            item[value] = item[key]["id"]
            del item[key]
        else:
            item[value] = None
            del item[key]
    json_fields = [
        "comments",
        "subscribers",
        "attachments",
        "labels",
        "subtasks",
        "projects",
        "memberships",
        "members",
    ]
    for field in json_fields:
        if item.get(field):
            item[f"{field}"] = item[field].get("nodes", [])

    return item


def _normalize_team(item: Dict[str, Any]) -> Dict[str, Any]:
    json_fields = ["memberships", "members", "projects"]
    for field in json_fields:
        if item.get(field):
            item[f"{field}"] = item[field].get("nodes", [])
    return item



"""
This module defines default settings for the Freshdesk integration.

It specifies a list of default endpoints to be used when interacting with the Freshdesk API,
covering common entities such as agents, companies, contacts, groups, roles, and tickets.
"""

# Define default endpoints for the Freshdesk API integration.
DEFAULT_ENDPOINTS = ["agents", "companies", "contacts", "groups", "roles", "tickets"]



from typing import Any, Dict, Iterable, Iterator

import dlt
import pendulum

from .helpers import _normalize_issue, _normalize_team, _paginate

ISSUES_QUERY = """
query Issues($cursor: String) {
  issues(first: 50, after: $cursor) {
    nodes {
      id
      title
      description
      createdAt
      updatedAt
      creator { id }
      assignee { id}
      state { id}
      labels { nodes { id } }
      cycle { id}
      project { id }
      subtasks: children { nodes { id title } }
      comments(first: 250) { nodes { id body } }
      priority
      attachments { nodes { id } }
      subscribers { nodes { id } }
    }
    pageInfo { hasNextPage endCursor }
  }
}
"""

PROJECTS_QUERY = """
query Projects($cursor: String) {
  projects(first: 50, after: $cursor) {
    nodes {
      id
      name
      description
      createdAt
      updatedAt
      health
      priority
      targetDate
      lead { id }
    }
    pageInfo { hasNextPage endCursor }
  }
}
"""

TEAMS_QUERY = """
query Teams($cursor: String) {
  teams(first: 50, after: $cursor) {
    nodes {
      id
      name
      key
      description
      updatedAt
      createdAt
      memberships { nodes { id } }
      members { nodes { id } }
      projects { nodes { id } }
    }
    pageInfo { hasNextPage endCursor }
  }
}
"""

USERS_QUERY = """
query Users($cursor: String) {
  users(first: 50, after: $cursor) {
    nodes {
      id
      name
      displayName
      email
      createdAt
      updatedAt
    }
    pageInfo { hasNextPage endCursor }
  }
}
"""


@dlt.source(name="linear", max_table_nesting=0)
def linear_source(
    api_key: str,
    start_date: pendulum.DateTime,
    end_date: pendulum.DateTime | None = None,
) -> Iterable[dlt.sources.DltResource]:
    @dlt.resource(name="issues", primary_key="id", write_disposition="merge")
    def issues(
        updated_at: dlt.sources.incremental[str] = dlt.sources.incremental(
            "updatedAt",
            initial_value=start_date.isoformat(),
            end_value=end_date.isoformat() if end_date else None,
            range_start="closed",
            range_end="closed",
        ),
    ) -> Iterator[Dict[str, Any]]:
        if updated_at.last_value:
            current_start_date = pendulum.parse(updated_at.last_value)
        else:
            current_start_date = pendulum.parse(start_date)

        if updated_at.end_value:
            current_end_date = pendulum.parse(updated_at.end_value)
        else:
            current_end_date = pendulum.now(tz="UTC")

        for item in _paginate(api_key, ISSUES_QUERY, "issues"):
            if pendulum.parse(item["updatedAt"]) >= current_start_date:
                if pendulum.parse(item["updatedAt"]) <= current_end_date:
                    yield _normalize_issue(item)

    @dlt.resource(name="projects", primary_key="id", write_disposition="merge")
    def projects(
        updated_at: dlt.sources.incremental[str] = dlt.sources.incremental(
            "updatedAt",
            initial_value=start_date.isoformat(),
            end_value=end_date.isoformat() if end_date else None,
            range_start="closed",
            range_end="closed",
        ),
    ) -> Iterator[Dict[str, Any]]:
        if updated_at.last_value:
            current_start_date = pendulum.parse(updated_at.last_value)
        else:
            current_start_date = pendulum.parse(start_date)

        if updated_at.end_value:
            current_end_date = pendulum.parse(updated_at.end_value)
        else:
            current_end_date = pendulum.now(tz="UTC")

        for item in _paginate(api_key, PROJECTS_QUERY, "projects"):
            if pendulum.parse(item["updatedAt"]) >= current_start_date:
                if pendulum.parse(item["updatedAt"]) <= current_end_date:
                    yield item

    @dlt.resource(name="teams", primary_key="id", write_disposition="merge")
    def teams(
        updated_at: dlt.sources.incremental[str] = dlt.sources.incremental(
            "updatedAt",
            initial_value=start_date.isoformat(),
            end_value=end_date.isoformat() if end_date else None,
            range_start="closed",
            range_end="closed",
        ),
    ) -> Iterator[Dict[str, Any]]:
        print(start_date)
        if updated_at.last_value:
            current_start_date = pendulum.parse(updated_at.last_value)
        else:
            current_start_date = pendulum.parse(start_date)
        print(current_start_date)

        if updated_at.end_value:
            current_end_date = pendulum.parse(updated_at.end_value)
        else:
            current_end_date = pendulum.now(tz="UTC")

        for item in _paginate(api_key, TEAMS_QUERY, "teams"):
            if pendulum.parse(item["updatedAt"]) >= current_start_date:
                if pendulum.parse(item["updatedAt"]) <= current_end_date:
                    yield _normalize_team(item)

    @dlt.resource(name="users", primary_key="id", write_disposition="merge")
    def users(
        updated_at: dlt.sources.incremental[str] = dlt.sources.incremental(
            "updatedAt",
            initial_value=start_date.isoformat(),
            end_value=end_date.isoformat() if end_date else None,
            range_start="closed",
            range_end="closed",
        ),
    ) -> Iterator[Dict[str, Any]]:
        if updated_at.last_value:
            current_start_date = pendulum.parse(updated_at.last_value)
        else:
            current_start_date = pendulum.parse(start_date)

        if updated_at.end_value:
            current_end_date = pendulum.parse(updated_at.end_value)
        else:
            current_end_date = pendulum.now(tz="UTC")

        for item in _paginate(api_key, USERS_QUERY, "users"):
            if pendulum.parse(item["updatedAt"]) >= current_start_date:
                if pendulum.parse(item["updatedAt"]) <= current_end_date:
                    yield item

    return issues, projects, teams, users



import abc
import base64
import csv
import datetime
import json
import os
import shutil
import struct
import tempfile
from urllib.parse import parse_qs, quote, urlparse

import dlt
import dlt.destinations.impl.filesystem.filesystem
from dlt.common.configuration.specs import AwsCredentials
from dlt.common.destination.capabilities import DestinationCapabilitiesContext
from dlt.common.schema import Schema
from dlt.common.storages.configuration import FileSystemCredentials
from dlt.destinations.impl.clickhouse.configuration import (
    ClickHouseCredentials,
)

from ingestr.src.errors import MissingValueError
from ingestr.src.loader import load_dlt_file


class GenericSqlDestination:
    def dlt_run_params(self, uri: str, table: str, **kwargs) -> dict:
        table_fields = table.split(".")
        if len(table_fields) != 2:
            raise ValueError("Table name must be in the format <schema>.<table>")

        res = {
            "dataset_name": table_fields[-2],
            "table_name": table_fields[-1],
        }

        return res

    def post_load(self):
        pass


class BigQueryDestination:
    def dlt_dest(self, uri: str, **kwargs):
        source_fields = urlparse(uri)
        source_params = parse_qs(source_fields.query)

        cred_path = source_params.get("credentials_path")
        credentials_base64 = source_params.get("credentials_base64")
        if not cred_path and not credentials_base64:
            raise ValueError(
                "credentials_path or credentials_base64 is required to connect BigQuery"
            )

        location = None
        if source_params.get("location"):
            loc_params = source_params.get("location", [])
            if len(loc_params) > 1:
                raise ValueError("Only one location is allowed")
            location = loc_params[0]

        credentials = {}
        if cred_path:
            with open(cred_path[0], "r") as f:
                credentials = json.load(f)
        elif credentials_base64:
            credentials = json.loads(
                base64.b64decode(credentials_base64[0]).decode("utf-8")
            )

        staging_bucket = kwargs.get("staging_bucket", None)
        if staging_bucket:
            if not staging_bucket.startswith("gs://"):
                raise ValueError("Staging bucket must start with gs://")

            os.environ["DESTINATION__FILESYSTEM__BUCKET_URL"] = staging_bucket
            os.environ["DESTINATION__FILESYSTEM__CREDENTIALS__PROJECT_ID"] = (
                credentials.get("project_id", None)
            )
            os.environ["DESTINATION__FILESYSTEM__CREDENTIALS__PRIVATE_KEY"] = (
                credentials.get("private_key", None)
            )
            os.environ["DESTINATION__FILESYSTEM__CREDENTIALS__CLIENT_EMAIL"] = (
                credentials.get("client_email", None)
            )

        project_id = None
        if source_fields.hostname:
            project_id = source_fields.hostname

        return dlt.destinations.bigquery(
            credentials=credentials,  # type: ignore
            location=location,  # type: ignore
            project_id=project_id,
            **kwargs,
        )

    def dlt_run_params(self, uri: str, table: str, **kwargs) -> dict:
        table_fields = table.split(".")
        if len(table_fields) != 2 and len(table_fields) != 3:
            raise ValueError(
                "Table name must be in the format <dataset>.<table> or <project>.<dataset>.<table>"
            )

        res = {
            "dataset_name": table_fields[-2],
            "table_name": table_fields[-1],
        }

        staging_bucket = kwargs.get("staging_bucket", None)
        if staging_bucket:
            res["staging"] = "filesystem"

        return res

    def post_load(self):
        pass


class CrateDBDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        uri = uri.replace("cratedb://", "postgres://")
        import dlt_cratedb.impl.cratedb.factory

        return dlt_cratedb.impl.cratedb.factory.cratedb(credentials=uri, **kwargs)


class PostgresDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.postgres(credentials=uri, **kwargs)


class SnowflakeDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.snowflake(credentials=uri, **kwargs)


class RedshiftDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.redshift(
            credentials=uri.replace("redshift://", "postgresql://"), **kwargs
        )


class DuckDBDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.duckdb(uri, **kwargs)


def handle_datetimeoffset(dto_value: bytes) -> datetime.datetime:
    # ref: https://github.com/mkleehammer/pyodbc/issues/134#issuecomment-281739794
    tup = struct.unpack(
        "<6hI2h", dto_value
    )  # e.g., (2017, 3, 16, 10, 35, 18, 500000000, -6, 0)
    return datetime.datetime(
        tup[0],
        tup[1],
        tup[2],
        tup[3],
        tup[4],
        tup[5],
        tup[6] // 1000,
        datetime.timezone(datetime.timedelta(hours=tup[7], minutes=tup[8])),
    )


# MSSQL_COPT_SS_ACCESS_TOKEN is a connection attribute used to pass
# an Azure Active Directory access token to the SQL Server ODBC driver.
MSSQL_COPT_SS_ACCESS_TOKEN = 1256


def serialize_azure_token(token):
    # https://github.com/mkleehammer/pyodbc/issues/228#issuecomment-494773723
    encoded = token.encode("utf_16_le")
    return struct.pack("<i", len(encoded)) + encoded


def build_mssql_dest():
    # https://github.com/bruin-data/ingestr/issues/293

    from dlt.destinations.impl.mssql.configuration import MsSqlClientConfiguration
    from dlt.destinations.impl.mssql.mssql import (
        HINT_TO_MSSQL_ATTR,
        MsSqlJobClient,
    )
    from dlt.destinations.impl.mssql.sql_client import (
        PyOdbcMsSqlClient,
    )

    class OdbcMsSqlClient(PyOdbcMsSqlClient):
        SKIP_CREDENTIALS = {"PWD", "AUTHENTICATION", "UID"}

        def open_connection(self):
            cfg = self.credentials._get_odbc_dsn_dict()
            if (
                cfg.get("AUTHENTICATION", "").strip().lower()
                != "activedirectoryaccesstoken"
            ):
                return super().open_connection()

            import pyodbc  # type: ignore

            dsn = ";".join(
                [f"{k}={v}" for k, v in cfg.items() if k not in self.SKIP_CREDENTIALS]
            )

            self._conn = pyodbc.connect(
                dsn,
                timeout=self.credentials.connect_timeout,
                attrs_before={
                    MSSQL_COPT_SS_ACCESS_TOKEN: serialize_azure_token(cfg["PWD"]),
                },
            )

            # https://github.com/mkleehammer/pyodbc/wiki/Using-an-Output-Converter-function
            self._conn.add_output_converter(-155, handle_datetimeoffset)
            self._conn.autocommit = True
            return self._conn

    class MsSqlClient(MsSqlJobClient):
        def __init__(
            self,
            schema: Schema,
            config: MsSqlClientConfiguration,
            capabilities: DestinationCapabilitiesContext,
        ) -> None:
            sql_client = OdbcMsSqlClient(
                config.normalize_dataset_name(schema),
                config.normalize_staging_dataset_name(schema),
                config.credentials,
                capabilities,
            )
            super(MsSqlJobClient, self).__init__(schema, config, sql_client)
            self.config: MsSqlClientConfiguration = config
            self.sql_client = sql_client
            self.active_hints = HINT_TO_MSSQL_ATTR if self.config.create_indexes else {}
            self.type_mapper = capabilities.get_type_mapper()

    class MsSqlDestImpl(dlt.destinations.mssql):
        @property
        def client_class(self):
            return MsSqlClient

    return MsSqlDestImpl


class MsSQLDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        cls = build_mssql_dest()
        return cls(credentials=uri, **kwargs)


class DatabricksDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.databricks(credentials=uri, **kwargs)


class SynapseDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.synapse(credentials=uri, **kwargs)


class CustomCsvDestination(dlt.destinations.filesystem):
    pass


class CsvDestination(GenericSqlDestination):
    temp_path: str
    actual_path: str
    uri: str
    dataset_name: str
    table_name: str

    def dlt_run_params(self, uri: str, table: str, **kwargs) -> dict:
        table_fields = table.split(".")
        if len(table_fields) != 2:
            raise ValueError("Table name must be in the format <schema>.<table>")

        res = {
            "dataset_name": table_fields[-2],
            "table_name": table_fields[-1],
        }

        self.dataset_name = res["dataset_name"]
        self.table_name = res["table_name"]
        self.uri = uri

        return res

    def dlt_dest(self, uri: str, **kwargs):
        if uri.startswith("csv://"):
            uri = uri.replace("csv://", "file://")

        temp_path = tempfile.mkdtemp()
        self.actual_path = uri
        self.temp_path = temp_path
        return CustomCsvDestination(bucket_url=f"file://{temp_path}", **kwargs)

    # I dislike this implementation quite a bit since it ties the implementation to some internal details on how dlt works
    # I would prefer a custom destination that allows me to do this easily but dlt seems to have a lot of internal details that are not documented
    # I tried to make it work with a nicer destination implementation but I couldn't, so I decided to go with this hack to experiment
    # if anyone has a better idea on how to do this, I am open to contributions or suggestions
    def post_load(self):
        def find_first_file(path):
            for entry in os.listdir(path):
                full_path = os.path.join(path, entry)
                if os.path.isfile(full_path):
                    return full_path

            return None

        def filter_keys(dictionary):
            return {
                key: value
                for key, value in dictionary.items()
                if not key.startswith("_dlt_")
            }

        first_file_path = find_first_file(
            f"{self.temp_path}/{self.dataset_name}/{self.table_name}"
        )

        output_path = self.uri.split("://")[1]
        if output_path.count("/") > 1:
            os.makedirs(os.path.dirname(output_path), exist_ok=True)

        with open(output_path, "w", newline="") as csv_file:
            csv_writer = None
            for row in load_dlt_file(first_file_path):
                row = filter_keys(row)
                if csv_writer is None:
                    csv_writer = csv.DictWriter(csv_file, fieldnames=row.keys())
                    csv_writer.writeheader()

                csv_writer.writerow(row)
        shutil.rmtree(self.temp_path)


class AthenaDestination:
    def dlt_dest(self, uri: str, **kwargs):
        encoded_uri = quote(uri, safe=":/?&=")
        source_fields = urlparse(encoded_uri)
        source_params = parse_qs(source_fields.query)

        bucket = source_params.get("bucket", [None])[0]
        if not bucket:
            raise ValueError("A bucket is required to connect to Athena.")

        if not bucket.startswith("s3://"):
            bucket = f"s3://{bucket}"

        bucket = bucket.rstrip("/")

        dest_table = kwargs.get("dest_table", None)
        if not dest_table:
            raise ValueError("A destination table is required to connect to Athena.")

        dest_table_fields = dest_table.split(".")
        if len(dest_table_fields) != 2:
            raise ValueError(
                f"Table name must be in the format <schema>.<table>, given: {dest_table}"
            )

        query_result_path = f"{bucket}/{dest_table_fields[0]}_staging/metadata"

        access_key_id = source_params.get("access_key_id", [None])[0]
        secret_access_key = source_params.get("secret_access_key", [None])[0]
        session_token = source_params.get("session_token", [None])[0]
        profile_name = source_params.get("profile", ["default"])[0]
        region_name = source_params.get("region_name", [None])[0]

        if not access_key_id and not secret_access_key:
            import botocore.session  # type: ignore

            session = botocore.session.Session(profile=profile_name)
            default = session.get_credentials()
            if not profile_name:
                raise ValueError(
                    "You have to either provide access_key_id and secret_access_key pair or a valid AWS profile name."
                )
            access_key_id = default.access_key
            secret_access_key = default.secret_key
            session_token = default.token
            if region_name is None:
                region_name = session.get_config_variable("region")

        if not region_name:
            raise ValueError("The region_name is required to connect to Athena.")

        os.environ["DESTINATION__BUCKET_URL"] = bucket
        if access_key_id and secret_access_key:
            os.environ["DESTINATION__CREDENTIALS__AWS_ACCESS_KEY_ID"] = access_key_id
            os.environ["DESTINATION__CREDENTIALS__AWS_SECRET_ACCESS_KEY"] = (
                secret_access_key
            )
        if session_token:
            os.environ["DESTINATION__CREDENTIALS__AWS_SESSION_TOKEN"] = session_token

        return dlt.destinations.athena(
            query_result_bucket=query_result_path,
            athena_work_group=source_params.get("workgroup", [None])[0],  # type: ignore
            credentials=AwsCredentials(
                aws_access_key_id=access_key_id,  # type: ignore
                aws_secret_access_key=secret_access_key,  # type: ignore
                aws_session_token=session_token,
                region_name=region_name,
            ),
            destination_name=bucket,
            force_iceberg=True,
        )

    def dlt_run_params(self, uri: str, table: str, **kwargs) -> dict:
        table_fields = table.split(".")
        if len(table_fields) != 2:
            raise ValueError("Table name must be in the format <schema>.<table>")
        return {
            "table_format": "iceberg",
            "dataset_name": table_fields[-2],
            "table_name": table_fields[-1],
        }

    def post_load(self):
        pass


class ClickhouseDestination:
    def dlt_dest(self, uri: str, **kwargs):
        parsed_uri = urlparse(uri)

        if "dest_table" in kwargs:
            table = kwargs["dest_table"]
            database = table.split(".")[0]
        else:
            database = parsed_uri.path.lstrip("/")

        username = parsed_uri.username
        if not username:
            raise ValueError(
                "A username is required to connect to the ClickHouse database."
            )

        password = parsed_uri.password
        if not password:
            raise ValueError(
                "A password is required to authenticate with the ClickHouse database."
            )

        host = parsed_uri.hostname
        if not host:
            raise ValueError(
                "The hostname or IP address of the ClickHouse server is required to establish a connection."
            )

        port = parsed_uri.port
        if not port:
            raise ValueError(
                "The TCP port of the ClickHouse server is required to establish a connection."
            )

        query_params = parse_qs(parsed_uri.query)
        secure = int(query_params["secure"][0]) if "secure" in query_params else 1

        http_port = (
            int(query_params["http_port"][0])
            if "http_port" in query_params
            else 8443
            if secure == 1
            else 8123
        )

        if secure not in (0, 1):
            raise ValueError(
                "Invalid value for secure. Set to `1` for a secure HTTPS connection or `0` for a non-secure HTTP connection."
            )

        credentials = ClickHouseCredentials(
            {
                "host": host,
                "port": port,
                "username": username,
                "password": password,
                "database": database,
                "http_port": http_port,
                "secure": secure,
            }
        )
        return dlt.destinations.clickhouse(credentials=credentials)

    def dlt_run_params(self, uri: str, table: str, **kwargs) -> dict:
        table_fields = table.split(".")
        if len(table_fields) != 2:
            raise ValueError("Table name must be in the format <schema>.<table>")
        return {
            "table_name": table_fields[-1],
        }

    def post_load(self):
        pass


class BlobFSClient(dlt.destinations.impl.filesystem.filesystem.FilesystemClient):
    @property
    def dataset_path(self):
        # override to remove dataset path
        return self.bucket_path


class BlobFS(dlt.destinations.filesystem):
    @property
    def client_class(self):
        return BlobFSClient


class SqliteDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.sqlalchemy(credentials=uri)

    def dlt_run_params(self, uri: str, table: str, **kwargs):
        return {
            # https://dlthub.com/docs/dlt-ecosystem/destinations/sqlalchemy#dataset-files
            "dataset_name": "main",
            "table_name": table,
        }


class MySqlDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.sqlalchemy(credentials=uri)

    def dlt_run_params(self, uri: str, table: str, **kwargs):
        parsed = urlparse(uri)
        database = parsed.path.lstrip("/")
        if not database:
            raise ValueError("You need to specify a database")
        return {
            "dataset_name": database,
            "table_name": table,
        }


class BlobStorageDestination(abc.ABC):
    @abc.abstractmethod
    def credentials(self, params: dict) -> FileSystemCredentials:
        """Build credentials for the blob storage destination."""
        pass

    @property
    @abc.abstractmethod
    def protocol(self) -> str:
        """The protocol used for the blob storage destination."""
        pass

    def dlt_dest(self, uri: str, **kwargs):
        parsed_uri = urlparse(uri)
        params = parse_qs(parsed_uri.query)
        creds = self.credentials(params)

        dest_table = kwargs["dest_table"]

        # only validate if dest_table is not a full URI
        if not parsed_uri.netloc:
            dest_table = self.validate_table(dest_table)

        table_parts = dest_table.split("/")

        if parsed_uri.path.strip("/"):
            path_parts = parsed_uri.path.strip("/ ").split("/")
            table_parts = path_parts + table_parts

        if parsed_uri.netloc:
            table_parts.insert(0, parsed_uri.netloc.strip())

        base_path = "/".join(table_parts[:-1])

        opts = {
            "bucket_url": f"{self.protocol}://{base_path}",
            "credentials": creds,
            # supresses dlt warnings about dataset name normalization.
            # we don't use dataset names in S3 so it's fine to disable this.
            "enable_dataset_name_normalization": False,
        }
        layout = params.get("layout", [None])[0]
        if layout is not None:
            opts["layout"] = layout

        return BlobFS(**opts)  # type: ignore

    def validate_table(self, table: str):
        table = table.strip("/ ")
        if len(table.split("/")) < 2:
            raise ValueError("Table name must be in the format {bucket-name}/{path}")
        return table

    def dlt_run_params(self, uri: str, table: str, **kwargs):
        table_parts = table.split("/")
        return {
            "table_name": table_parts[-1].strip(),
        }

    def post_load(self) -> None:
        pass


class S3Destination(BlobStorageDestination):
    @property
    def protocol(self) -> str:
        return "s3"

    def credentials(self, params: dict) -> FileSystemCredentials:
        access_key_id = params.get("access_key_id", [None])[0]
        if access_key_id is None:
            raise MissingValueError("access_key_id", "S3")

        secret_access_key = params.get("secret_access_key", [None])[0]
        if secret_access_key is None:
            raise MissingValueError("secret_access_key", "S3")

        endpoint_url = params.get("endpoint_url", [None])[0]
        if endpoint_url is not None:
            parsed_endpoint = urlparse(endpoint_url)
            if not parsed_endpoint.scheme or not parsed_endpoint.netloc:
                raise ValueError("Invalid endpoint_url. Must be a valid URL.")

        return AwsCredentials(
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            endpoint_url=endpoint_url,
        )


class GCSDestination(BlobStorageDestination):
    @property
    def protocol(self) -> str:
        return "gs"

    def credentials(self, params: dict) -> FileSystemCredentials:
        """Builds GCS credentials from the provided parameters."""
        credentials_path = params.get("credentials_path")
        credentials_base64 = params.get("credentials_base64")
        credentials_available = any(
            map(
                lambda x: x is not None,
                [credentials_path, credentials_base64],
            )
        )
        if credentials_available is False:
            raise MissingValueError("credentials_path or credentials_base64", "GCS")

        credentials = None
        if credentials_path:
            with open(credentials_path[0], "r") as f:
                credentials = json.load(f)
        else:
            credentials = json.loads(base64.b64decode(credentials_base64[0]).decode())  # type: ignore

        return credentials



"""
This module handles how credentials are read in dlt sources
"""

from typing import ClassVar, List, Union

import dlt
from dlt.common.configuration import configspec
from dlt.common.configuration.specs import CredentialsConfiguration
from dlt.common.typing import TSecretValue


@configspec
class ZendeskCredentialsBase(CredentialsConfiguration):
    """
    The Base version of all the ZendeskCredential classes.
    """

    subdomain: str = dlt.config.value
    __config_gen_annotations__: ClassVar[List[str]] = []


@configspec
class ZendeskCredentialsEmailPass(ZendeskCredentialsBase):
    """
    This class is used to store credentials for Email + Password Authentication
    """

    email: str = dlt.config.value
    password: TSecretValue = dlt.secrets.value


@configspec
class ZendeskCredentialsOAuth(ZendeskCredentialsBase):
    """
    This class is used to store credentials for OAuth Token Authentication
    """

    oauth_token: TSecretValue = dlt.secrets.value


@configspec
class ZendeskCredentialsToken(ZendeskCredentialsBase):
    """
    This class is used to store credentials for Token Authentication
    """

    email: str = dlt.config.value
    token: TSecretValue = dlt.secrets.value


TZendeskCredentials = Union[
    ZendeskCredentialsEmailPass, ZendeskCredentialsToken, ZendeskCredentialsOAuth
]



"""Freshdesk Client for making authenticated requests"""

import logging
import time
from typing import Any, Dict, Iterable, Optional

from dlt.common.typing import TDataItem
from dlt.sources.helpers import requests


class FreshdeskClient:
    """
    Client for making authenticated requests to the Freshdesk API. It incorporates API requests with
    rate limit and pagination.

    Attributes:
        api_key (str): The API key used for authenticating requests to the Freshdesk API.
        domain (str): The Freshdesk domain specific to the user, used in constructing the base URL.
        base_url (str): The base URL constructed from the domain, targeting the Freshdesk API v2.
    """

    def __init__(self, api_key: str, domain: str):
        # Initialize the FreshdeskClient instance with API key and domain.
        # The API key is used for authentication with the Freshdesk API.
        # The domain specifies the unique Freshdesk domain of the user.

        # Store the API key provided during initialization.
        self.api_key = api_key
        # Store the Freshdesk domain provided during initialization.
        self.domain = domain

        # Construct the base URL for the API requests.
        # This URL is formed by appending the domain to the standard Freshdesk API base URL format.
        # All API requests will use this base URL as their starting point.
        self.base_url = f"https://{domain}.freshdesk.com/api/v2"

    def _request_with_rate_limit(self, url: str, **kwargs: Any) -> requests.Response:
        """
        Handles rate limits in HTTP requests and ensures
        that the client doesn't exceed the limit set by the server.
        """

        while True:
            try:
                response = requests.get(url, **kwargs, auth=(self.api_key, "X"))
                response.raise_for_status()

                return response
            except requests.HTTPError as e:
                if e.response.status_code == 429:
                    # Get the 'Retry-After' header to know how long to wait
                    # Fallback to 60 seconds if header is missing
                    seconds_to_wait = int(e.response.headers.get("Retry-After", 60))
                    # Log a warning message
                    logging.warning(
                        "Rate limited. Waiting to retry after: %s secs", seconds_to_wait
                    )

                    # Wait for the specified number of seconds before retrying
                    time.sleep(seconds_to_wait)
                else:
                    # If the error is not a rate limit (429), raise the exception to be
                    # handled elsewhere or stop execution
                    raise

    def paginated_response(
        self,
        endpoint: str,
        per_page: int,
        updated_at: Optional[str] = None,
    ) -> Iterable[TDataItem]:
        """
        Fetches a paginated response from a specified endpoint.

        This method will continuously fetch data from the given endpoint,
        page by page, until no more data is available or until it reaches data
        updated at the specified timestamp.
        """
        page = 1
        while True:
            # Construct the URL for the specific endpoint
            url = f"{self.base_url}/{endpoint}"

            params: Dict[str, Any] = {"per_page": per_page, "page": page}

            # Implement date range splitting logic here, if applicable
            if endpoint in ["tickets", "contacts"]:
                param_key = (
                    "updated_since" if endpoint == "tickets" else "_updated_since"
                )
                if updated_at:
                    params[param_key] = updated_at

            # Handle requests with rate-limiting
            # A maximum of 300 pages (30000 tickets) will be returned.
            response = self._request_with_rate_limit(url, params=params)
            data = response.json()

            if not data:
                break  # Stop if no data or max page limit reached
            yield data
            page += 1



from typing import Any, Iterator, Optional

import dlt
from dlt.common.pendulum import pendulum
from dlt.common.time import ensure_pendulum_datetime
from dlt.common.typing import TAnyDateTime

from ingestr.src.frankfurter.helpers import get_path_with_retry


@dlt.source(
    name="frankfurter",
    max_table_nesting=0,
)
def frankfurter_source(
    start_date: TAnyDateTime,
    end_date: TAnyDateTime,
    base_currency: str,
) -> Any:
    """
    A dlt source for the frankfurter.dev API. It groups several resources (in this case frankfurter.dev API endpoints) containing
    various types of data: currencies, latest rates, historical rates.
    """
    date_time = dlt.sources.incremental(
        "date",
        initial_value=start_date,
        end_value=end_date,
        range_start="closed",
        range_end="closed",
    )

    return (
        currencies(),
        latest(base_currency=base_currency),
        exchange_rates(
            start_date=date_time, end_date=end_date, base_currency=base_currency
        ),
    )


@dlt.resource(
    write_disposition="replace",
    columns={
        "currency_code": {"data_type": "text"},
        "currency_name": {"data_type": "text"},
    },
)
def currencies() -> Iterator[dict]:
    """
    Yields each currency as a separate row with two columns: currency_code and currency_name.
    """
    # Retrieve the list of currencies from the API
    currencies_data = get_path_with_retry("currencies")

    for currency_code, currency_name in currencies_data.items():
        yield {"currency_code": currency_code, "currency_name": currency_name}


@dlt.resource(
    write_disposition="merge",
    columns={
        "date": {"data_type": "text"},
        "currency_code": {"data_type": "text"},
        "rate": {"data_type": "double"},
        "base_currency": {"data_type": "text"},
    },
    primary_key=["date", "currency_code", "base_currency"],
)
def latest(base_currency: Optional[str] = "") -> Iterator[dict]:
    """
    Fetches the latest exchange rates and yields them as rows.
    """
    # Base URL
    url = "latest?"

    if base_currency:
        url += f"base={base_currency}"

    # Fetch data
    data = get_path_with_retry(url)

    # Extract rates and base currency
    rates = data["rates"]
    date = pendulum.parse(data["date"])

    # Add the base currency with a rate of 1.0
    yield {
        "date": date,
        "currency_code": base_currency,
        "rate": 1.0,
        "base_currency": base_currency,
    }

    # Add all currencies and their rates
    for currency_code, rate in rates.items():
        yield {
            "date": date,
            "currency_code": currency_code,
            "rate": rate,
            "base_currency": base_currency,
        }


@dlt.resource(
    write_disposition="merge",
    columns={
        "date": {"data_type": "text"},
        "currency_code": {"data_type": "text"},
        "rate": {"data_type": "double"},
        "base_currency": {"data_type": "text"},
    },
    primary_key=("date", "currency_code", "base_currency"),
)
def exchange_rates(
    end_date: TAnyDateTime,
    start_date: dlt.sources.incremental[TAnyDateTime] = dlt.sources.incremental("date"),
    base_currency: Optional[str] = "",
) -> Iterator[dict]:
    """
    Fetches exchange rates for a specified date range.
    If only start_date is provided, fetches data until now.
    If both start_date and end_date are provided, fetches data for each day in the range.
    """
    # Ensure start_date.last_value is a pendulum.DateTime object
    start_date_obj = ensure_pendulum_datetime(start_date.last_value)  # type: ignore
    start_date_str = start_date_obj.format("YYYY-MM-DD")

    # Ensure end_date is a pendulum.DateTime object
    end_date_obj = ensure_pendulum_datetime(end_date)
    end_date_str = end_date_obj.format("YYYY-MM-DD")

    # Compose the URL
    url = f"{start_date_str}..{end_date_str}?"

    if base_currency:
        url += f"base={base_currency}"

    # Fetch data from the API
    data = get_path_with_retry(url)

    # Extract base currency and rates from the API response
    rates = data["rates"]

    # Iterate over the rates dictionary (one entry per date)
    for date, daily_rates in rates.items():
        formatted_date = pendulum.parse(date)

        # Add the base currency with a rate of 1.0
        yield {
            "date": formatted_date,
            "currency_code": base_currency,
            "rate": 1.0,
            "base_currency": base_currency,
        }

        # Add all other currencies and their rates
        for currency_code, rate in daily_rates.items():
            yield {
                "date": formatted_date,
                "currency_code": currency_code,
                "rate": rate,
                "base_currency": base_currency,
            }



"""Slack source helpers."""

from typing import Any, Generator, Iterable, List, Optional
from urllib.parse import urljoin

import pendulum
from dlt.common.time import ensure_pendulum_datetime
from dlt.common.typing import Dict, TAnyDateTime, TDataItem
from dlt.sources.helpers import requests
from jsonpath_ng.ext import parse  # type: ignore

from .settings import MAX_PAGE_SIZE, SLACK_API_URL


class SlackApiException(Exception):
    """Slack api exception."""


class PaidOnlyException(SlackApiException):
    """Slack api exception."""


def extract_jsonpath(
    expression: str,
    json_data: TDataItem,
) -> Generator[Any, None, None]:
    """Extract records from an input based on a JSONPath expression."""
    if not expression:
        yield json_data
        return

    jsonpath = parse(expression)

    for match in jsonpath.find(json_data):
        yield match.value


def update_jsonpath(expression: str, json_data: TDataItem, value: Any) -> Any:
    """Update a record in an input based on a JSONPath expression."""
    jsonpath = parse(expression)
    return jsonpath.update_or_create(json_data, value)


def ensure_dt_type(dt: TAnyDateTime, to_ts: bool = False) -> Any:
    """Converts a datetime to a pendulum datetime or timestamp.
    Args:
        dt: The datetime to convert.
        to_ts: Whether to convert to a timestamp or not.
    Returns:
        Any: The converted datetime or timestamp.
    """
    if dt is None:
        return None
    out_dt = ensure_pendulum_datetime(dt)
    if to_ts:
        return out_dt.timestamp()
    return out_dt


class SlackAPI:
    """
    A Slack API client that can be used to get pages of data from Slack.
    """

    def __init__(
        self,
        access_token: str,
        page_size: int = MAX_PAGE_SIZE,
    ) -> None:
        """
        Args:
            access_token: The private app password to the app on your shop.
            page_size: The max number of items to fetch per page. Defaults to 1000.
        """
        self.access_token = access_token
        self.page_size = page_size

    @property
    def headers(self) -> Dict[str, str]:
        """Generate the headers to use for the request."""
        return {"Authorization": f"Bearer {self.access_token}"}

    def parameters(
        self, params: Optional[Dict[str, Any]] = None, next_cursor: str = None
    ) -> Dict[str, str]:
        """
        Generate the query parameters to use for the request.

        Args:
            params: The query parameters to include in the request.
            next_cursor: The cursor to use to get the next page of results.

        Returns:
            The query parameters to use for the request.
        """
        params = params or {}
        params["limit"] = self.page_size
        if next_cursor:
            params["cursor"] = next_cursor
        return params

    def url(self, resource: str) -> str:
        """
        Generate the URL to use for the request.

        Args:
            resource: The resource to get pages for (e.g. conversations.list).
        """
        return urljoin(SLACK_API_URL, resource)

    def _get_next_cursor(self, response: Dict[str, Any]) -> Any:
        """
        Get the next cursor from the response.

        Args:
            response: The response from the Slack API.
        """
        cursor_jsonpath = "$.response_metadata.next_cursor"
        return next(extract_jsonpath(cursor_jsonpath, response), None)

    def _convert_datetime_fields(
        self, item: Dict[str, Any], datetime_fields: List[str]
    ) -> Dict[str, Any]:
        """Convert timestamp fields in the item to pendulum datetime objects.

        The item is modified in place.

        Args:
            item: The item to convert
            datetime_fields: List of fields to convert to pendulum datetime objects.

        Returns:
            The same data item (for convenience)
        """
        if not datetime_fields:
            return item

        for field in datetime_fields:
            if timestamp := next(extract_jsonpath(field, item), None):
                if isinstance(timestamp, str):
                    timestamp = float(timestamp)
                if timestamp > 1e10:
                    timestamp = timestamp / 1000
                pendulum_dt = pendulum.from_timestamp(timestamp)
                item = update_jsonpath(field, item, pendulum_dt)
        return item

    def get_pages(
        self,
        resource: str,
        response_path: str = None,
        params: Dict[str, Any] = None,
        datetime_fields: List[str] = None,
        context: Dict[str, Any] = None,
    ) -> Iterable[TDataItem]:
        """Get all pages from slack using requests.
        Iterates through all pages and yield each page items.\

        Args:
            resource: The resource to get pages for (e.g. conversations.list).
            response_path: The path to the list of items in the response JSON.
            params: Query params to include in the request.
            datetime_fields: List of fields to convert to pendulum datetime objects.
            context: Additional context to add to each item.

        Yields:
            List of data items from the page
        """
        has_next_page = True
        next_cursor = None

        # Iterate through all pages
        while has_next_page:
            # Make the request
            response = requests.get(
                url=self.url(resource),
                headers=self.headers,
                params=self.parameters(params or {}, next_cursor),
            )
            json_response = response.json()

            # Stop if there was an error
            if not json_response.get("ok"):
                has_next_page = False
                error = json_response.get("error")
                if error == "paid_only":
                    raise PaidOnlyException(
                        "This resource is just available on paid accounts."
                    )
                else:
                    raise SlackApiException(error)

            # Yield the page converting datetime fields
            output = []
            for item in extract_jsonpath(response_path, json_response):
                item = self._convert_datetime_fields(item, datetime_fields)
                item.update(context or {})
                output.append(item)
            yield output

            # Get the next cursor
            next_cursor = self._get_next_cursor(json_response)
            if not next_cursor:
                has_next_page = False



from typing import Iterable

import dlt
import pendulum
from dlt.common.time import ensure_pendulum_datetime
from dlt.common.typing import TDataItem
from dlt.sources import DltResource
from dlt.sources.helpers import requests


@dlt.source(name="pinterest", max_table_nesting=0)
def pinterest_source(
    start_date: pendulum.DateTime,
    access_token: str,
    page_size: int = 200,
    end_date: pendulum.DateTime | None = None,
) -> Iterable[DltResource]:
    session = requests.Session()
    session.headers.update({"Authorization": f"Bearer {access_token}"})
    base_url = "https://api.pinterest.com/v5"

    def fetch_data(
        endpoint: str,
        start_dt: pendulum.DateTime,
        end_dt: pendulum.DateTime,
    ) -> Iterable[TDataItem]:
        url = f"{base_url}/{endpoint}"
        params = {"page_size": page_size}
        bookmark = None
        while True:
            if bookmark:
                params["bookmark"] = bookmark

            resp = session.get(url, params=params)
            resp.raise_for_status()
            data = resp.json()
            items = data.get("items") or []

            for item in items:
                item_created = ensure_pendulum_datetime(item["created_at"])
                if item_created <= start_dt:
                    continue
                if item_created > end_dt:
                    continue
                item["created_at"] = item_created
                yield item

            bookmark = data.get("bookmark")
            if not bookmark:
                break

    @dlt.resource(write_disposition="merge", primary_key="id")
    def pins(
        datetime=dlt.sources.incremental(
            "created_at",
            initial_value=start_date,
            end_value=end_date,
        ),
    ) -> Iterable[TDataItem]:
        _start_date = datetime.last_value or start_date
        if end_date is None:
            _end_date = pendulum.now("UTC")
        else:
            _end_date = datetime.end_value
        yield from fetch_data("pins", _start_date, _end_date)

    @dlt.resource(write_disposition="merge", primary_key="id")
    def boards(
        datetime=dlt.sources.incremental(
            "created_at",
            initial_value=start_date,
            end_value=end_date,
        ),
    ) -> Iterable[TDataItem]:
        _start_date = datetime.last_value or start_date
        if end_date is None:
            _end_date = pendulum.now("UTC")
        else:
            _end_date = datetime.end_value
        yield from fetch_data("boards", _start_date, _end_date)

    return pins, boards



"""Facebook ads source settings and constants"""

from typing import Any, Callable, Dict, Iterator, Literal

from dlt.common.schema.typing import TTableSchemaColumns
from facebook_business.adobjects.abstractobject import AbstractObject

TFbMethod = Callable[..., Iterator[AbstractObject]]


DEFAULT_FIELDS = (
    "id",
    "updated_time",
    "created_time",
    "name",
    "status",
    "effective_status",
)

DEFAULT_CAMPAIGN_FIELDS = DEFAULT_FIELDS + (
    "objective",
    "start_time",
    "stop_time",
    "daily_budget",
    "lifetime_budget",
)

DEFAULT_AD_FIELDS = DEFAULT_FIELDS + (
    "adset_id",
    "campaign_id",
    "creative",
    "targeting",
    "tracking_specs",
    "conversion_specs",
)

DEFAULT_ADSET_FIELDS = DEFAULT_FIELDS + (
    "campaign_id",
    "start_time",
    "end_time",
    "daily_budget",
    "lifetime_budget",
    "optimization_goal",
    "promoted_object",
    "billing_event",
    "bid_amount",
    "bid_strategy",
    "targeting",
)

DEFAULT_ADCREATIVE_FIELDS = (
    "id",
    "name",
    "status",
    "thumbnail_url",
    "object_story_spec",
    "effective_object_story_id",
    "call_to_action_type",
    "object_type",
    "template_url",
    "url_tags",
    "instagram_actor_id",
    "product_set_id",
)

DEFAULT_LEAD_FIELDS = (
    "id",
    "created_time",
    "ad_id",
    "ad_name",
    "adset_id",
    "adset_name",
    "campaign_id",
    "campaign_name",
    "form_id",
    "field_data",
)

DEFAULT_INSIGHT_FIELDS = (
    "campaign_id",
    "adset_id",
    "ad_id",
    "date_start",
    "date_stop",
    "reach",
    "impressions",
    "frequency",
    "clicks",
    "unique_clicks",
    "ctr",
    "unique_ctr",
    "cpc",
    "cpm",
    "cpp",
    "spend",
    "actions",
    "action_values",
    "cost_per_action_type",
    "website_ctr",
    "account_currency",
    "ad_click_actions",
    "ad_name",
    "adset_name",
    "campaign_name",
    "country",
    "dma",
    "full_view_impressions",
    "full_view_reach",
    "inline_link_click_ctr",
    "outbound_clicks",
    "reach",
    "social_spend",
    "spend",
    "website_ctr",
    "conversions",
    "video_thruplay_watched_actions",
)

TInsightsLevels = Literal["account", "campaign", "adset", "ad"]

INSIGHTS_PRIMARY_KEY = ("campaign_id", "adset_id", "ad_id", "date_start")

ALL_STATES = {
    "effective_status": [
        "ACTIVE",
        "PAUSED",
        "DELETED",
        "PENDING_REVIEW",
        "DISAPPROVED",
        "PREAPPROVED",
        "PENDING_BILLING_INFO",
        "CAMPAIGN_PAUSED",
        "ARCHIVED",
        "ADSET_PAUSED",
    ]
}

TInsightsBreakdownOptions = Literal[
    "ads_insights",
    "ads_insights_age_and_gender",
    "ads_insights_country",
    "ads_insights_platform_and_device",
    "ads_insights_region",
    "ads_insights_dma",
    "ads_insights_hourly_advertiser",
]

ALL_ACTION_ATTRIBUTION_WINDOWS = (
    "1d_click",
    "7d_click",
    "28d_click",
    "1d_view",
    "7d_view",
    "28d_view",
)

ALL_ACTION_BREAKDOWNS = ("action_type", "action_target_id", "action_destination")

INSIGHTS_BREAKDOWNS_OPTIONS: Dict[TInsightsBreakdownOptions, Any] = {
    "ads_insights": {"breakdowns": (), "fields": ()},
    "ads_insights_age_and_gender": {
        "breakdowns": ("age", "gender"),
        "fields": ("age", "gender"),
    },
    "ads_insights_country": {"breakdowns": ("country",), "fields": ("country",)},
    "ads_insights_platform_and_device": {
        "breakdowns": ("publisher_platform", "platform_position", "impression_device"),
        "fields": ("publisher_platform", "platform_position", "impression_device"),
    },
    "ads_insights_region": {"breakdowns": ("region",), "fields": ("region",)},
    "ads_insights_dma": {"breakdowns": ("dma",), "fields": ("dma",)},
    "ads_insights_hourly_advertiser": {
        "breakdowns": ("hourly_stats_aggregated_by_advertiser_time_zone",),
        "fields": ("hourly_stats_aggregated_by_advertiser_time_zone",),
    },
}

INSIGHT_FIELDS_TYPES: TTableSchemaColumns = {
    "campaign_id": {"data_type": "bigint"},
    "adset_id": {"data_type": "bigint"},
    "ad_id": {"data_type": "bigint"},
    "date_start": {"data_type": "timestamp"},
    "date_stop": {"data_type": "timestamp"},
    "reach": {"data_type": "bigint"},
    "impressions": {"data_type": "bigint"},
    "frequency": {"data_type": "decimal"},
    "clicks": {"data_type": "bigint"},
    "unique_clicks": {"data_type": "bigint"},
    "ctr": {"data_type": "decimal"},
    "unique_ctr": {"data_type": "decimal"},
    "cpc": {"data_type": "decimal"},
    "cpm": {"data_type": "decimal"},
    "cpp": {"data_type": "decimal"},
    "spend": {"data_type": "decimal"},
}

INVALID_INSIGHTS_FIELDS = [
    "impression_device",
    "publisher_platform",
    "platform_position",
    "age",
    "gender",
    "country",
    "placement",
    "region",
    "dma",
    "hourly_stats_aggregated_by_advertiser_time_zone",
]

FACEBOOK_INSIGHTS_RETENTION_PERIOD = 37  # months



"""Contains helper functions to extract data from spreadsheet API"""

from typing import Any, List, Tuple

from dlt.common.exceptions import MissingDependencyException
from dlt.common.typing import DictStrAny
from dlt.sources.credentials import GcpCredentials, GcpOAuthCredentials
from dlt.sources.helpers.requests.retry import DEFAULT_RETRY_STATUS
from tenacity import retry, retry_if_exception, stop_after_attempt, wait_exponential

from .data_processing import ParsedRange, trim_range_top_left

try:
    from apiclient.discovery import Resource, build
except ImportError:
    raise MissingDependencyException("Google API Client", ["google-api-python-client"])


def is_retry_status_code(exception: BaseException) -> bool:
    """Retry condition on HttpError"""
    from googleapiclient.errors import HttpError  # type: ignore

    # print(f"RETRY ON {str(HttpError)} = {isinstance(exception, HttpError) and exception.resp.status in DEFAULT_RETRY_STATUS}")
    # if isinstance(exception, HttpError):
    #     print(exception.resp.status)
    #     print(DEFAULT_RETRY_STATUS)
    return (
        isinstance(exception, HttpError)
        and exception.resp.status in DEFAULT_RETRY_STATUS
    )


retry_deco = retry(
    # Retry if it's a rate limit error (HTTP 429)
    retry=retry_if_exception(is_retry_status_code),
    # Use exponential backoff for the waiting time between retries, starting with 5 seconds
    wait=wait_exponential(multiplier=1.5, min=5, max=120),
    # Stop retrying after 10 attempts
    stop=stop_after_attempt(10),
    # Print out the retrying details
    reraise=True,
)


def api_auth(credentials: GcpCredentials, max_api_retries: int) -> Resource:
    """
    Uses GCP credentials to authenticate with Google Sheets API.

    Args:
        credentials (GcpCredentials): Credentials needed to log in to GCP.
        max_api_retries (int): Max number of retires to google sheets API. Actual behavior is internal to google client.

    Returns:
        Resource: Object needed to make API calls to Google Sheets API.
    """
    if isinstance(credentials, GcpOAuthCredentials):
        credentials.auth("https://www.googleapis.com/auth/spreadsheets.readonly")
    # Build the service object for Google sheets api.
    service = build(
        "sheets",
        "v4",
        credentials=credentials.to_native_credentials(),
        num_retries=max_api_retries,
    )
    return service


@retry_deco
def get_meta_for_ranges(
    service: Resource, spreadsheet_id: str, range_names: List[str]
) -> Any:
    """Retrieves `spreadsheet_id` cell metadata for `range_names`"""
    return (
        service.spreadsheets()
        .get(
            spreadsheetId=spreadsheet_id,
            ranges=range_names,
            includeGridData=True,
        )
        .execute()
    )


@retry_deco
def get_known_range_names(
    spreadsheet_id: str, service: Resource
) -> Tuple[List[str], List[str], str]:
    """
    Retrieves spreadsheet metadata and extracts a list of sheet names and named ranges

    Args:
        spreadsheet_id (str): The ID of the spreadsheet.
        service (Resource): Resource object used to make API calls to Google Sheets API.

    Returns:
        Tuple[List[str], List[str], str] sheet names, named ranges, spreadheet title
    """
    metadata = service.spreadsheets().get(spreadsheetId=spreadsheet_id).execute()
    sheet_names: List[str] = [s["properties"]["title"] for s in metadata["sheets"]]
    named_ranges: List[str] = [r["name"] for r in metadata.get("namedRanges", {})]
    title: str = metadata["properties"]["title"]
    return sheet_names, named_ranges, title


@retry_deco
def get_data_for_ranges(
    service: Resource, spreadsheet_id: str, range_names: List[str]
) -> List[Tuple[str, ParsedRange, ParsedRange, List[List[Any]]]]:
    """
    Calls Google Sheets API to get data in a batch. This is the most efficient way to get data for multiple ranges inside a spreadsheet.

    Args:
        service (Resource): Object to make API calls to Google Sheets.
        spreadsheet_id (str): The ID of the spreadsheet.
        range_names (List[str]): List of range names.

    Returns:
        List[DictStrAny]: A list of ranges with data in the same order as `range_names`
    """
    range_batch_resp = (
        service.spreadsheets()
        .values()
        .batchGet(
            spreadsheetId=spreadsheet_id,
            ranges=range_names,
            # un formatted returns typed values
            valueRenderOption="UNFORMATTED_VALUE",
            # will return formatted dates as a serial number
            dateTimeRenderOption="SERIAL_NUMBER",
        )
        .execute()
    )
    # if there are not ranges to be loaded, there's no "valueRanges"
    range_batch: List[DictStrAny] = range_batch_resp.get("valueRanges", [])
    # trim the empty top rows and columns from the left
    rv = []
    for name, range_ in zip(range_names, range_batch):
        parsed_range = ParsedRange.parse_range(range_["range"])
        values: List[List[Any]] = range_.get("values", None)
        if values:
            parsed_range, values = trim_range_top_left(parsed_range, values)
        # create a new range to get first two rows
        meta_range = parsed_range._replace(end_row=parsed_range.start_row + 1)
        # print(f"{name}:{parsed_range}:{meta_range}")
        rv.append((name, parsed_range, meta_range, values))
    return rv



"""A source that extracts data from Notion API"""

from typing import Dict, Iterator, List, Optional

import dlt
from dlt.sources import DltResource

from .helpers.client import NotionClient
from .helpers.database import NotionDatabase


@dlt.source(max_table_nesting=1)
def notion_databases(
    database_ids: Optional[List[Dict[str, str]]] = None,
    api_key: str = dlt.secrets.value,
) -> Iterator[DltResource]:
    """
    Retrieves data from Notion databases.

    Args:
        database_ids (List[Dict[str, str]], optional): A list of dictionaries
            each containing a database id and a name.
            Defaults to None. If None, the function will generate all databases
            in the workspace that are accessible to the integration.
        api_key (str): The Notion API secret key.

    Yields:
        DltResource: Data resources from Notion databases.
    """
    notion_client = NotionClient(api_key)

    if database_ids is None:
        search_results = notion_client.search(
            filter_criteria={"value": "database", "property": "object"}
        )
        database_ids = [
            {"id": result["id"], "use_name": result["title"][0]["plain_text"]}
            for result in search_results
        ]

    for database in database_ids:
        if "use_name" not in database:
            # Fetch the database details from Notion
            details = notion_client.get_database(database["id"])

            # Extract the name/title from the details
            database["use_name"] = details["title"][0]["plain_text"]

        notion_database = NotionDatabase(database["id"], notion_client)
        yield dlt.resource(  # type: ignore
            notion_database.query(),
            primary_key="id",
            name=database["use_name"],
            write_disposition="replace",
        )



import time
from typing import Any, Dict, Iterator, Optional

import pendulum

from ingestr.src.http_client import create_client


class ZoomClient:
    """Minimal Zoom API client supporting Server-to-Server OAuth."""

    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        account_id: Optional[str] = None,
    ) -> None:
        self.client_id = client_id
        self.client_secret = client_secret
        self.account_id = account_id
        self.token_expires_at: float = 0
        self.base_url = "https://api.zoom.us/v2"
        self.session = create_client()
        self._refresh_access_token()

    def _refresh_access_token(self) -> None:
        token_url = "https://zoom.us/oauth/token"
        auth = (self.client_id, self.client_secret)
        resp = self.session.post(
            token_url,
            params={"grant_type": "account_credentials", "account_id": self.account_id},
            auth=auth,
        )
        resp.raise_for_status()
        data = resp.json()
        self.access_token = data.get("access_token")
        self.token_expires_at = time.time() + data.get("expires_in", 3600)

    def _ensure_token(self) -> None:
        if self.access_token is None or self.token_expires_at <= time.time():
            self._refresh_access_token()

    def _headers(self) -> Dict[str, str]:
        self._ensure_token()
        return {
            "Authorization": f"Bearer {self.access_token}",
            "Accept": "application/json",
        }

    def get_users(self) -> Iterator[Dict[str, Any]]:
        url = f"{self.base_url}/users"

        params = {"page_size": 1000}
        while True:
            response = self.session.get(url, headers=self._headers(), params=params)
            response.raise_for_status()
            data = response.json()
            for user in data.get("users", []):
                yield user
            token = data.get("next_page_token")
            if not token:
                break
            params["next_page_token"] = token

    # https://developers.zoom.us/docs/api/rest/reference/zoom-api/methods/#operation/meetings
    def get_meetings(
        self, user_id: str, params: Dict[str, Any]
    ) -> Iterator[Dict[str, Any]]:
        url = f"{self.base_url}/users/{user_id}/meetings"
        while True:
            response = self.session.get(url, headers=self._headers(), params=params)
            response.raise_for_status()
            data = response.json()
            for item in data.get("meetings", []):
                item["zoom_user_id"] = user_id
                yield item
            token = data.get("next_page_token")
            if not token:
                break
            params["next_page_token"] = token

    # https://developers.zoom.us/docs/api/rest/reference/zoom-api/methods/#operation/reportMeetingParticipants
    def get_participants(
        self,
        meeting_id: str,
        params: Dict[str, Any],
        start_date: pendulum.DateTime,
        end_date: pendulum.DateTime,
    ) -> Iterator[Dict[str, Any]]:
        url = f"{self.base_url}/report/meetings/{meeting_id}/participants"
        while True:
            response = self.session.get(url, headers=self._headers(), params=params)
            response.raise_for_status()
            data = response.json()
            for item in data.get("participants", []):
                join_time = pendulum.parse(item["join_time"])
                if join_time >= start_date and join_time <= end_date:
                    yield item
            token = data.get("next_page_token")
            if not token:
                break
            params["next_page_token"] = token



from typing import Optional

from dlt.common.runtime.collector import Collector
from rich.status import Status


class SpinnerCollector(Collector):
    status: Status
    current_step: str
    started: bool

    def __init__(self) -> None:
        self.status = Status("Ingesting data...", spinner="dots")
        self.started = False

    def update(
        self,
        name: str,
        inc: int = 1,
        total: Optional[int] = None,
        message: Optional[str] = None,  # type: ignore
        label: str = "",
        **kwargs,
    ) -> None:
        self.status.update(self.current_step)

    def _start(self, step: str) -> None:
        self.current_step = self.__step_to_label(step)
        self.status.start()

    def __step_to_label(self, step: str) -> str:
        verb = step.split(" ")[0].lower()
        if verb.startswith("normalize"):
            return "Normalizing the data"
        elif verb.startswith("load"):
            return "Loading the data to the destination"
        elif verb.startswith("extract"):
            return "Extracting the data from the source"

        return f"{verb.capitalize()} the data"

    def _stop(self) -> None:
        self.status.stop()



from typing import Iterable, Iterator

import dlt
import pendulum
from dlt.sources import DltResource

from .helpers import SolidgateClient

COLUMN_HINTS = {
    "subscriptions": {
        "id": {"data_type": "text", "nullable": False, "primary_key": True},
        "created_at": {"data_type": "timestamp", "partition": True},
        "status": {"data_type": "text"},
        "started_at": {"data_type": "timestamp"},
        "updated_at": {"data_type": "timestamp"},
        "expired_at": {"data_type": "timestamp"},
        "next_charge_at": {"data_type": "timestamp"},
        "payment_type": {"data_type": "text"},
        "trial": {"data_type": "bool"},
        "cancelled_at": {"data_type": "timestamp"},
        "cancellation_requested_at": {"data_type": "timestamp"},
        "cancel_code": {"data_type": "text"},
        "cancel_message": {"data_type": "text"},
        "customer": {"data_type": "json"},
        "product": {"data_type": "json"},
        "invoices": {"data_type": "json"},
    },
    "apm_orders": {
        "order_id": {"data_type": "text", "nullable": False, "primary_key": True},
        "created_at": {"data_type": "timestamp", "partition": True},
        "updated_at": {"data_type": "timestamp"},
        "order_description": {"data_type": "text"},
        "method": {"data_type": "text"},
        "amount": {"data_type": "bigint"},
        "currency": {"data_type": "text"},
        "processing_amount": {"data_type": "bigint"},
        "processing_currency": {"data_type": "text"},
        "status": {"data_type": "text"},
        "customer_account_id": {"data_type": "text"},
        "customer_email": {"data_type": "text"},
        "ip_address": {"data_type": "text"},
        "geo_country": {"data_type": "text"},
        "error_code": {"data_type": "text"},
        "transactions": {"data_type": "json"},
        "order_metadata": {"data_type": "json"},
    },
    "card_orders": {
        "order_id": {"data_type": "text", "nullable": False, "primary_key": True},
        "created_at": {"data_type": "timestamp", "partition": True},
        "updated_at": {"data_type": "timestamp"},
        "order_description": {"data_type": "text"},
        "psp_order_id": {"data_type": "text"},
        "provider_payment_id": {"data_type": "text"},
        "amount": {"data_type": "bigint"},
        "currency": {"data_type": "text"},
        "processing_amount": {"data_type": "bigint"},
        "processing_currency": {"data_type": "text"},
        "status": {"data_type": "text"},
        "payment_type": {"data_type": "text"},
        "type": {"data_type": "text"},
        "is_secured": {"data_type": "bool"},
        "routing": {"data_type": "json"},
        "customer_account_id": {"data_type": "text"},
        "customer_email": {"data_type": "text"},
        "customer_first_name": {"data_type": "text"},
        "customer_last_name": {"data_type": "text"},
        "ip_address": {"data_type": "text"},
        "mid": {"data_type": "text"},
        "traffic_source": {"data_type": "text"},
        "platform": {"data_type": "text"},
        "geo_country": {"data_type": "text"},
        "error_code": {"data_type": "text"},
        "transactions": {"data_type": "json"},
        "order_metadata": {"data_type": "json"},
        "fraudulent": {"data_type": "bool"},
    },
    "financial_entries": {
        "id": {
            "data_type": "text",
            "nullable": False,
            "primary_key": True,
        },
        "order_id": {"data_type": "text"},
        "external_psp_order_id": {"data_type": "text"},
        "created_at": {"data_type": "timestamp", "partition": True},
        "transaction_datetime_provider": {"data_type": "timestamp"},
        "transaction_datetime_utc": {"data_type": "timestamp"},
        "accounting_date": {"data_type": "date"},
        "amount": {"data_type": "double"},
        "amount_in_major_units": {"data_type": "double"},
        "currency": {"data_type": "text"},
        "currency_minor_units": {"data_type": "bigint"},
        "payout_amount": {"data_type": "double"},
        "payout_amount_in_major_units": {"data_type": "double"},
        "payout_currency": {"data_type": "text"},
        "payout_currency_minor_units": {"data_type": "bigint"},
        "record_type_key": {"data_type": "text"},
        "provider": {"data_type": "text"},
        "payment_method": {"data_type": "text"},
        "card_brand": {"data_type": "text"},
        "geo_country": {"data_type": "text"},
        "issuing_country": {"data_type": "text"},
        "transaction_id": {"data_type": "text"},
        "chargeback_id": {"data_type": "text"},
        "legal_entity": {"data_type": "text"},
    },
}


@dlt.source(max_table_nesting=0)
def solidgate_source(
    start_date: pendulum.DateTime,
    end_date: pendulum.DateTime | None,
    public_key: str,
    secret_key: str,
) -> Iterable[DltResource]:
    solidgate_client = SolidgateClient(public_key, secret_key)

    @dlt.resource(
        name="subscriptions",
        write_disposition="merge",
        primary_key="id",
        columns=COLUMN_HINTS["subscriptions"],  # type: ignore
    )
    def fetch_all_subscriptions(
        dateTime=dlt.sources.incremental(
            "updated_at",
            initial_value=start_date,
            end_value=end_date,
            range_start="closed",
            range_end="closed",
        ),
    ) -> Iterator[dict]:
        path = "subscriptions"
        if dateTime.end_value is None:
            end_dt = pendulum.now(tz="UTC")
        else:
            end_dt = dateTime.end_value

        start_dt = dateTime.last_value
        yield solidgate_client.fetch_data(path, date_from=start_dt, date_to=end_dt)

    @dlt.resource(
        name="apm_orders",
        write_disposition="merge",
        primary_key="order_id",
        columns=COLUMN_HINTS["apm_orders"],  # type: ignore
    )
    def fetch_apm_orders(
        dateTime=dlt.sources.incremental(
            "updated_at",
            initial_value=start_date,
            end_value=end_date,
            range_start="closed",
            range_end="closed",
        ),
    ) -> Iterator[dict]:
        path = "apm-orders"
        if dateTime.end_value is None:
            end_dt = pendulum.now(tz="UTC")
        else:
            end_dt = dateTime.end_value

        start_dt = dateTime.last_value
        yield solidgate_client.fetch_data(path, date_from=start_dt, date_to=end_dt)

    @dlt.resource(
        name="card_orders",
        write_disposition="merge",
        primary_key="order_id",
        columns=COLUMN_HINTS["card_orders"],  # type: ignore
    )
    def fetch_card_orders(
        dateTime=dlt.sources.incremental(
            "updated_at",
            initial_value=start_date,
            end_value=end_date,
            range_start="closed",
            range_end="closed",
        ),
    ) -> Iterator[dict]:
        path = "card-orders"
        if dateTime.end_value is None:
            end_dt = pendulum.now(tz="UTC")
        else:
            end_dt = dateTime.end_value

        start_dt = dateTime.last_value
        yield solidgate_client.fetch_data(path, date_from=start_dt, date_to=end_dt)

    @dlt.resource(
        name="financial_entries",
        write_disposition="merge",
        primary_key="id",
        columns=COLUMN_HINTS["financial_entries"],  # type: ignore
    )
    def fetch_financial_entries(
        dateTime=dlt.sources.incremental(
            "created_at",
            initial_value=start_date,
            end_value=end_date,
            range_start="closed",
            range_end="closed",
        ),
    ):
        if dateTime.end_value is None:
            end_date = pendulum.now(tz="UTC")
        else:
            end_date = dateTime.end_value

        start_date = dateTime.last_value
        yield solidgate_client.fetch_financial_entry_data(start_date, end_date)

    return (
        fetch_all_subscriptions,
        fetch_apm_orders,
        fetch_card_orders,
        fetch_financial_entries,
    )



"""Hubspot source settings and constants"""

from dlt.common import pendulum

STARTDATE = pendulum.datetime(year=2000, month=1, day=1)

CRM_CONTACTS_ENDPOINT = (
    "/crm/v3/objects/contacts?associations=companies,deals,products,tickets,quotes"
)
CRM_COMPANIES_ENDPOINT = "/crm/v3/objects/companies?associations=products"
CRM_DEALS_ENDPOINT = (
    "/crm/v3/objects/deals?associations=companies,contacts,products,tickets,quotes"
)
CRM_PRODUCTS_ENDPOINT = (
    "/crm/v3/objects/products?associations=companies,contacts,deals,tickets,quotes"
)
CRM_TICKETS_ENDPOINT = (
    "/crm/v3/objects/tickets?associations=companies,contacts,deals,products,quotes"
)
CRM_QUOTES_ENDPOINT = (
    "/crm/v3/objects/quotes?associations=companies,contacts,deals,products,tickets"
)
CRM_SCHEMAS_ENDPOINT = "/crm/v3/schemas"

CRM_OBJECT_ENDPOINTS = {
    "contact": CRM_CONTACTS_ENDPOINT,
    "company": CRM_COMPANIES_ENDPOINT,
    "deal": CRM_DEALS_ENDPOINT,
    "product": CRM_PRODUCTS_ENDPOINT,
    "ticket": CRM_TICKETS_ENDPOINT,
    "quote": CRM_QUOTES_ENDPOINT,
}

WEB_ANALYTICS_EVENTS_ENDPOINT = "/events/v3/events?objectType={objectType}&objectId={objectId}&occurredAfter={occurredAfter}&occurredBefore={occurredBefore}&sort=-occurredAt"

OBJECT_TYPE_SINGULAR = {
    "companies": "company",
    "contacts": "contact",
    "deals": "deal",
    "tickets": "ticket",
    "products": "product",
    "quotes": "quote",
}

OBJECT_TYPE_PLURAL = {v: k for k, v in OBJECT_TYPE_SINGULAR.items()}

DEFAULT_DEAL_PROPS = [
    "amount",
    "closedate",
    "createdate",
    "dealname",
    "dealstage",
    "hs_lastmodifieddate",
    "hs_object_id",
    "pipeline",
]

DEFAULT_COMPANY_PROPS = [
    "createdate",
    "domain",
    "hs_lastmodifieddate",
    "hs_object_id",
    "name",
]

DEFAULT_CONTACT_PROPS = [
    "createdate",
    "email",
    "firstname",
    "hs_object_id",
    "lastmodifieddate",
    "lastname",
]

DEFAULT_TICKET_PROPS = [
    "createdate",
    "content",
    "hs_lastmodifieddate",
    "hs_object_id",
    "hs_pipeline",
    "hs_pipeline_stage",
    "hs_ticket_category",
    "hs_ticket_priority",
    "subject",
]

DEFAULT_PRODUCT_PROPS = [
    "createdate",
    "description",
    "hs_lastmodifieddate",
    "hs_object_id",
    "name",
    "price",
]

DEFAULT_QUOTE_PROPS = [
    "hs_createdate",
    "hs_expiration_date",
    "hs_lastmodifieddate",
    "hs_object_id",
    "hs_public_url_key",
    "hs_status",
    "hs_title",
]

ALL = ("ALL",)



"""Mongo database source helpers"""

import json
from itertools import islice
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)

import dlt
from bson.decimal128 import Decimal128
from bson.objectid import ObjectId
from bson.regex import Regex
from bson.timestamp import Timestamp
from dlt.common import logger
from dlt.common.configuration.specs import BaseConfiguration, configspec
from dlt.common.data_writers import TDataItemFormat
from dlt.common.time import ensure_pendulum_datetime
from dlt.common.typing import TDataItem
from dlt.common.utils import map_nested_in_place
from pendulum import _datetime
from pymongo import ASCENDING, DESCENDING, MongoClient
from pymongo.collection import Collection
from pymongo.cursor import Cursor
from pymongo.helpers_shared import _fields_list_to_dict

if TYPE_CHECKING:
    TMongoClient = MongoClient[Any]
    TCollection = Collection[Any]
    TCursor = Cursor[Any]
else:
    TMongoClient = Any
    TCollection = Any
    TCursor = Any

try:
    import pymongoarrow  # type: ignore

    PYMONGOARROW_AVAILABLE = True
except ImportError:
    PYMONGOARROW_AVAILABLE = False


class CollectionLoader:
    def __init__(
        self,
        client: TMongoClient,
        collection: TCollection,
        chunk_size: int,
        incremental: Optional[dlt.sources.incremental[Any]] = None,
    ) -> None:
        self.client = client
        self.collection = collection
        self.incremental = incremental
        self.chunk_size = chunk_size

        if incremental:
            self.cursor_field = incremental.cursor_path
            self.last_value = incremental.last_value
        else:
            self.cursor_column = None
            self.last_value = None

    @property
    def _sort_op(self) -> List[Optional[Tuple[str, int]]]:
        if not self.incremental or not self.last_value:
            return []

        if (
            self.incremental.row_order == "asc"
            and self.incremental.last_value_func is max
        ) or (
            self.incremental.row_order == "desc"
            and self.incremental.last_value_func is min
        ):
            return [(self.cursor_field, ASCENDING)]

        elif (
            self.incremental.row_order == "asc"
            and self.incremental.last_value_func is min
        ) or (
            self.incremental.row_order == "desc"
            and self.incremental.last_value_func is max
        ):
            return [(self.cursor_field, DESCENDING)]

        return []

    @property
    def _filter_op(self) -> Dict[str, Any]:
        """Build a filtering operator.

        Includes a field and the filtering condition for it.

        Returns:
            Dict[str, Any]: A dictionary with the filter operator.
        """
        if not (self.incremental and self.last_value):
            return {}

        filt = {}
        if self.incremental.last_value_func is max:
            filt = {self.cursor_field: {"$gte": self.last_value}}
            if self.incremental.end_value:
                filt[self.cursor_field]["$lt"] = self.incremental.end_value

        elif self.incremental.last_value_func is min:
            filt = {self.cursor_field: {"$lte": self.last_value}}
            if self.incremental.end_value:
                filt[self.cursor_field]["$gt"] = self.incremental.end_value

        return filt

    def _projection_op(
        self, projection: Optional[Union[Mapping[str, Any], Iterable[str]]]
    ) -> Optional[Dict[str, Any]]:
        """Build a projection operator.

        Args:
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): A tuple of fields to include or a dict specifying fields to include or exclude.
            The incremental `primary_key` needs to be handle differently for inclusion
            and exclusion projections.

        Returns:
            Tuple[str, ...] | Dict[str, Any]: A tuple or dictionary with the projection operator.
        """
        if projection is None:
            return None

        projection_dict = dict(_fields_list_to_dict(projection, "projection"))

        if self.incremental:
            # this is an inclusion projection
            if any(v == 1 for v in projection_dict.values()):
                # ensure primary_key is included
                projection_dict.update(m={self.incremental.primary_key: 1})
            # this is an exclusion projection
            else:
                try:
                    # ensure primary_key isn't excluded
                    projection_dict.pop(self.incremental.primary_key)  # type: ignore
                except KeyError:
                    pass  # primary_key was properly not included in exclusion projection
                else:
                    dlt.common.logger.warn(
                        f"Primary key `{self.incremental.primary_key}` was removed from exclusion projection"
                    )

        return projection_dict

    def _limit(self, cursor: Cursor, limit: Optional[int] = None) -> TCursor:  # type: ignore
        """Apply a limit to the cursor, if needed.

        Args:
            cursor (Cursor): The cursor to apply the limit.
            limit (Optional[int]): The number of documents to load.

        Returns:
            Cursor: The cursor with the limit applied (if given).
        """
        if limit not in (0, None):
            if self.incremental is None or self.incremental.last_value_func is None:
                logger.warning(
                    "Using limit without ordering - results may be inconsistent."
                )

            cursor = cursor.limit(abs(limit))

        return cursor

    def load_documents(
        self,
        filter_: Dict[str, Any],
        limit: Optional[int] = None,
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
    ) -> Iterator[TDataItem]:
        """Construct the query and load the documents from the collection.

        Args:
            filter_ (Dict[str, Any]): The filter to apply to the collection.
            limit (Optional[int]): The number of documents to load.
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.

        Yields:
            Iterator[TDataItem]: An iterator of the loaded documents.
        """
        filter_op = self._filter_op
        _raise_if_intersection(filter_op, filter_)
        filter_op.update(filter_)

        projection_op = self._projection_op(projection)

        cursor = self.collection.find(filter=filter_op, projection=projection_op)
        if self._sort_op:
            cursor = cursor.sort(self._sort_op)

        cursor = self._limit(cursor, limit)

        while docs_slice := list(islice(cursor, self.chunk_size)):
            res = map_nested_in_place(convert_mongo_objs, docs_slice)
            if len(res) > 0 and "_id" in res[0] and isinstance(res[0]["_id"], dict):
                yield dlt.mark.with_hints(
                    res,
                    dlt.mark.make_hints(columns={"_id": {"data_type": "json"} }),
                )
            else:
                yield res


class CollectionLoaderParallel(CollectionLoader):
    def _get_document_count(self) -> int:
        return self.collection.count_documents(filter=self._filter_op)

    def _create_batches(self, limit: Optional[int] = None) -> List[Dict[str, int]]:
        doc_count = self._get_document_count()
        if limit:
            doc_count = min(doc_count, abs(limit))

        batches = []
        left_to_load = doc_count

        for sk in range(0, doc_count, self.chunk_size):
            batches.append(dict(skip=sk, limit=min(self.chunk_size, left_to_load)))
            left_to_load -= self.chunk_size

        return batches

    def _get_cursor(
        self,
        filter_: Dict[str, Any],
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
    ) -> TCursor:
        """Get a reading cursor for the collection.

        Args:
            filter_ (Dict[str, Any]): The filter to apply to the collection.
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.

        Returns:
            Cursor: The cursor for the collection.
        """
        filter_op = self._filter_op
        _raise_if_intersection(filter_op, filter_)
        filter_op.update(filter_)

        projection_op = self._projection_op(projection)

        cursor = self.collection.find(filter=filter_op, projection=projection_op)
        if self._sort_op:
            cursor = cursor.sort(self._sort_op)

        return cursor

    @dlt.defer
    def _run_batch(self, cursor: TCursor, batch: Dict[str, int]) -> TDataItem:
        cursor = cursor.clone()

        data = []
        for document in cursor.skip(batch["skip"]).limit(batch["limit"]):
            data.append(map_nested_in_place(convert_mongo_objs, document))

        return data

    def _get_all_batches(
        self,
        filter_: Dict[str, Any],
        limit: Optional[int] = None,
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
    ) -> Iterator[TDataItem]:
        """Load all documents from the collection in parallel batches.

        Args:
            filter_ (Dict[str, Any]): The filter to apply to the collection.
            limit (Optional[int]): The maximum number of documents to load.
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.

        Yields:
            Iterator[TDataItem]: An iterator of the loaded documents.
        """
        batches = self._create_batches(limit=limit)
        cursor = self._get_cursor(filter_=filter_, projection=projection)

        for batch in batches:
            yield self._run_batch(cursor=cursor, batch=batch)

    def load_documents(
        self,
        filter_: Dict[str, Any],
        limit: Optional[int] = None,
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
    ) -> Iterator[TDataItem]:
        """Load documents from the collection in parallel.

        Args:
            filter_ (Dict[str, Any]): The filter to apply to the collection.
            limit (Optional[int]): The number of documents to load.
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.

        Yields:
            Iterator[TDataItem]: An iterator of the loaded documents.
        """
        for document in self._get_all_batches(
            limit=limit, filter_=filter_, projection=projection
        ):
            yield document


class CollectionArrowLoader(CollectionLoader):
    """
    Mongo DB collection loader, which uses
    Apache Arrow for data processing.
    """

    def load_documents(
        self,
        filter_: Dict[str, Any],
        limit: Optional[int] = None,
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
        pymongoarrow_schema: Any = None,
    ) -> Iterator[Any]:
        """
        Load documents from the collection in Apache Arrow format.

        Args:
            filter_ (Dict[str, Any]): The filter to apply to the collection.
            limit (Optional[int]): The number of documents to load.
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.
            pymongoarrow_schema (Any): The mapping of field types to convert BSON to Arrow.

        Yields:
            Iterator[Any]: An iterator of the loaded documents.
        """
        from pymongoarrow.context import PyMongoArrowContext  # type: ignore
        from pymongoarrow.lib import process_bson_stream  # type: ignore

        filter_op = self._filter_op
        _raise_if_intersection(filter_op, filter_)
        filter_op.update(filter_)

        projection_op = self._projection_op(projection)

        # NOTE the `filter_op` isn't passed
        cursor = self.collection.find_raw_batches(
            filter_, batch_size=self.chunk_size, projection=projection_op
        )
        if self._sort_op:
            cursor = cursor.sort(self._sort_op)  # type: ignore

        cursor = self._limit(cursor, limit)  # type: ignore

        context = PyMongoArrowContext.from_schema(
            schema=pymongoarrow_schema, codec_options=self.collection.codec_options
        )
        for batch in cursor:
            process_bson_stream(batch, context)
            table = context.finish()
            yield convert_arrow_columns(table)


class CollectionArrowLoaderParallel(CollectionLoaderParallel):
    """
    Mongo DB collection parallel loader, which uses
    Apache Arrow for data processing.
    """

    def load_documents(
        self,
        filter_: Dict[str, Any],
        limit: Optional[int] = None,
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
        pymongoarrow_schema: Any = None,
    ) -> Iterator[TDataItem]:
        """Load documents from the collection in parallel.

        Args:
            filter_ (Dict[str, Any]): The filter to apply to the collection.
            limit (Optional[int]): The number of documents to load.
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.
            pymongoarrow_schema (Any): The mapping of field types to convert BSON to Arrow.

        Yields:
            Iterator[TDataItem]: An iterator of the loaded documents.
        """
        yield from self._get_all_batches(
            limit=limit,
            filter_=filter_,
            projection=projection,
            pymongoarrow_schema=pymongoarrow_schema,
        )

    def _get_all_batches(
        self,
        filter_: Dict[str, Any],
        limit: Optional[int] = None,
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
        pymongoarrow_schema: Any = None,
    ) -> Iterator[TDataItem]:
        """Load all documents from the collection in parallel batches.

        Args:
            filter_ (Dict[str, Any]): The filter to apply to the collection.
            limit (Optional[int]): The maximum number of documents to load.
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.
            pymongoarrow_schema (Any): The mapping of field types to convert BSON to Arrow.

        Yields:
            Iterator[TDataItem]: An iterator of the loaded documents.
        """
        batches = self._create_batches(limit=limit)
        cursor = self._get_cursor(filter_=filter_, projection=projection)
        for batch in batches:
            yield self._run_batch(
                cursor=cursor,
                batch=batch,
                pymongoarrow_schema=pymongoarrow_schema,
            )

    def _get_cursor(
        self,
        filter_: Dict[str, Any],
        projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = None,
    ) -> TCursor:
        """Get a reading cursor for the collection.

        Args:
            filter_ (Dict[str, Any]): The filter to apply to the collection.
            projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.

        Returns:
            Cursor: The cursor for the collection.
        """
        filter_op = self._filter_op
        _raise_if_intersection(filter_op, filter_)
        filter_op.update(filter_)

        projection_op = self._projection_op(projection)

        cursor = self.collection.find_raw_batches(
            filter=filter_op, batch_size=self.chunk_size, projection=projection_op
        )
        if self._sort_op:
            cursor = cursor.sort(self._sort_op)  # type: ignore

        return cursor

    @dlt.defer
    def _run_batch(
        self,
        cursor: TCursor,
        batch: Dict[str, int],
        pymongoarrow_schema: Any = None,
    ) -> TDataItem:
        from pymongoarrow.context import PyMongoArrowContext
        from pymongoarrow.lib import process_bson_stream

        cursor = cursor.clone()

        context = PyMongoArrowContext.from_schema(
            schema=pymongoarrow_schema, codec_options=self.collection.codec_options
        )
        for chunk in cursor.skip(batch["skip"]).limit(batch["limit"]):
            process_bson_stream(chunk, context)
            table = context.finish()
            yield convert_arrow_columns(table)


def collection_documents(
    client: TMongoClient,
    collection: TCollection,
    filter_: Dict[str, Any],
    projection: Union[Dict[str, Any], List[str]],
    pymongoarrow_schema: "pymongoarrow.schema.Schema",
    incremental: Optional[dlt.sources.incremental[Any]] = None,
    parallel: bool = False,
    limit: Optional[int] = None,
    chunk_size: Optional[int] = 10000,
    data_item_format: Optional[TDataItemFormat] = "object",
) -> Iterator[TDataItem]:
    """
    A DLT source which loads data from a Mongo database using PyMongo.
    Resources are automatically created for the collection.

    Args:
        client (MongoClient): The PyMongo client `pymongo.MongoClient` instance.
        collection (Collection): The collection `pymongo.collection.Collection` to load.
        filter_ (Dict[str, Any]): The filter to apply to the collection.
        projection (Optional[Union[Mapping[str, Any], Iterable[str]]]): The projection to select fields to create the Cursor.
            when loading the collection. Supported inputs:
                include (list) - ["year", "title"]
                include (dict) - {"year": True, "title": True}
                exclude (dict) - {"released": False, "runtime": False}
            Note: Can't mix include and exclude statements '{"title": True, "released": False}`
        pymongoarrow_schema (pymongoarrow.schema.Schema): The mapping of field types to convert BSON to Arrow.
        incremental (Optional[dlt.sources.incremental[Any]]): The incremental configuration.
        parallel (bool): Option to enable parallel loading for the collection. Default is False.
        limit (Optional[int]): The maximum number of documents to load.
        chunk_size (Optional[int]): The number of documents to load in each batch.
        data_item_format (Optional[TDataItemFormat]): The data format to use for loading.
            Supported formats:
                object - Python objects (dicts, lists).
                arrow - Apache Arrow tables.

    Returns:
        Iterable[DltResource]: A list of DLT resources for each collection to be loaded.
    """
    if data_item_format == "arrow" and not PYMONGOARROW_AVAILABLE:
        dlt.common.logger.warn(
            "'pymongoarrow' is not installed; falling back to standard MongoDB CollectionLoader."
        )
        data_item_format = "object"

    if data_item_format != "arrow" and pymongoarrow_schema:
        dlt.common.logger.warn(
            "Received value for `pymongoarrow_schema`, but `data_item_format=='object'` "
            "Use `data_item_format=='arrow'` to enforce schema."
        )

    if data_item_format == "arrow" and pymongoarrow_schema and projection:
        dlt.common.logger.warn(
            "Received values for both `pymongoarrow_schema` and `projection`. Since both "
            "create a projection to select fields, `projection` will be ignored."
        )

    if parallel:
        if data_item_format == "arrow":
            LoaderClass = CollectionArrowLoaderParallel
        else:
            LoaderClass = CollectionLoaderParallel  # type: ignore
    else:
        if data_item_format == "arrow":
            LoaderClass = CollectionArrowLoader  # type: ignore
        else:
            LoaderClass = CollectionLoader  # type: ignore
    
    loader = LoaderClass(
        client, collection, incremental=incremental, chunk_size=chunk_size
    )
    if isinstance(loader, (CollectionArrowLoader, CollectionArrowLoaderParallel)):
        yield from loader.load_documents(
            limit=limit,
            filter_=filter_,
            projection=projection,
            pymongoarrow_schema=pymongoarrow_schema,
        )
    else:
        yield from loader.load_documents(
            limit=limit, filter_=filter_, projection=projection
        )


def convert_mongo_objs(value: Any) -> Any:
    """MongoDB to dlt type conversion when using Python loaders.

    Notes:
        The method `ObjectId.__str__()` creates a hexstring using `binascii.hexlify(__id).decode()`

    """
    if isinstance(value, (ObjectId, Decimal128)):
        return str(value)
    if isinstance(value, _datetime.datetime):
        return ensure_pendulum_datetime(value)
    if isinstance(value, Regex):
        return value.try_compile().pattern
    if isinstance(value, Timestamp):
        date = value.as_datetime()
        return ensure_pendulum_datetime(date)

    return value


def convert_arrow_columns(table: Any) -> Any:
    """Convert the given table columns to Python types.

    Notes:
        Calling str() matches the `convert_mongo_obs()` used in non-arrow code.
        Pymongoarrow converts ObjectId to `fixed_size_binary[12]`, which can't be
        converted to a string as a vectorized operation because it contains ASCII characters.

        Instead, you need to loop over values using:
        ```python
        pyarrow.array([v.as_buffer().hex() for v in object_id_array], type=pyarrow.string())
        # pymongoarrow simplifies this by allowing this syntax
        [str(v) for v in object_id_array]
        ```

    Args:
        table (pyarrow.lib.Table): The table to convert.

    Returns:
        pyarrow.lib.Table: The table with the columns converted.
    """
    from dlt.common.libs.pyarrow import pyarrow
    from pymongoarrow.types import (  # type: ignore
        _is_binary,
        _is_code,
        _is_decimal128,
        _is_objectid,
    )

    for i, field in enumerate(table.schema):
        if _is_objectid(field.type) or _is_decimal128(field.type):
            col_values = [str(value) for value in table[field.name]]
            table = table.set_column(
                i,
                pyarrow.field(field.name, pyarrow.string()),
                pyarrow.array(col_values, type=pyarrow.string()),
            )
        else:
            type_ = None
            if _is_binary(field.type):
                type_ = pyarrow.binary()
            elif _is_code(field.type):
                type_ = pyarrow.string()

            if type_:
                col_values = [value.as_py() for value in table[field.name]]
                table = table.set_column(
                    i,
                    pyarrow.field(field.name, type_),
                    pyarrow.array(col_values, type=type_),
                )
    return table


def client_from_credentials(connection_url: str) -> TMongoClient:
    client: TMongoClient = MongoClient(
        connection_url, uuidRepresentation="standard", tz_aware=True
    )
    return client


def _raise_if_intersection(filter1: Dict[str, Any], filter2: Dict[str, Any]) -> None:
    """
    Raise an exception, if the given filters'
    fields are intersecting.

    Args:
        filter1 (Dict[str, Any]): The first filter.
        filter2 (Dict[str, Any]): The second filter.
    """
    field_inter = filter1.keys() & filter2.keys()
    for field in field_inter:
        if filter1[field].keys() & filter2[field].keys():
            str_repr = str({field: filter1[field]})
            raise ValueError(
                (
                    f"Filtering operator {str_repr} is already used by the "
                    "incremental and can't be used in the filter."
                )
            )


@configspec
class MongoDbCollectionConfiguration(BaseConfiguration):
    incremental: Optional[dlt.sources.incremental] = None  # type: ignore[type-arg]


@configspec
class MongoDbCollectionResourceConfiguration(BaseConfiguration):
    connection_url: dlt.TSecretValue = dlt.secrets.value
    database: Optional[str] = dlt.config.value
    collection: str = dlt.config.value
    incremental: Optional[dlt.sources.incremental] = None  # type: ignore[type-arg]
    write_disposition: Optional[str] = dlt.config.value
    parallel: Optional[bool] = False
    projection: Optional[Union[Mapping[str, Any], Iterable[str]]] = dlt.config.value


__source_name__ = "mongodb"
